"""
The MIT License (MIT)

Copyright (c) 2021-present UnrealFar & TheGenocides
Copyright (c) 2023-present Sengolda

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import base64
import hashlib
import hmac
import json
import logging
import time
import threading
import datetime
from urllib.parse import urlparse
from asyncio import iscoroutinefunction
from http import HTTPStatus
from typing import Callable, List, Optional, Union, Any
from flask import Flask, request

from .paginations import MessagePagination
from .attachments import CTA, CustomProfile, File, Geo, Poll, QuickReply
from .enums import ReplySetting, SpaceState, Granularity, JobType, JobStatus
from .constants import TWEET_EXPANSION, USER_FIELD, MEDIA_FIELD, PLACE_FIELD, POLL_FIELD, TWEET_FIELD
from .errors import pytwotException, UnKnownSpaceState
from .http import HTTPClient
from .message import DirectMessage, WelcomeMessage, WelcomeMessageRule
from .space import Space
from .stream import Stream
from .tweet import Tweet
from .user import User, ClientAccount
from .environment import Environment, Webhook
from .dataclass import Location, Trend
from .list import List as TwitterList
from .type import ID
from .compliance import Job

__all__ = ("Client",)

_log = logging.getLogger(__name__)


class Client:
    """Represents a twitter-api client for twitter api version 1.1 and 2 interface.

    Parameters
    ------------
    bearer_token: Optional[:class:`str`]
        The Bearer Token of the app. Uses for twitter version 2 endpoints.
    consumer_key: Optional[:class:`str`]
        The Consumer Key of the app. Users for twitter version 1.1 endpoints.
    consumer_secret: Optional[:class:`str`]
        The Consumer Key Secret of the app. Users for twitter version 1.1 endpoints.
    access_token: Optional[:class:`str`]
        The Access Token of the app. Users for twitter version 1.1 endpoints.
    access_token_secret: Optional[:class:`str`]
        The Access Token Secret of the app. Users for twitter version 1.1 endpoints.
    stream: Optional[Stream]
        The client's stream. Must be an instance of :class:`Stream`.
    callback_url: Optional[:class:`str`]
        The oauth callback url, default to None. Makes sure the callback url is the same as the one in your application auth-settings or else it can't create and interact with oauth related methods.
    client_id: Optional[:class:`str`]
        The client's OAuth 2.0 Client ID from keys and tokens page.
    client_secret: Optional[:class:`str`]
        The client's OAuth 2.0 Client Secret from keys and tokens page.
    use_bearer_only: bool
        Indicates to only use bearer token for all methods. This mean the client is now a twitter-api-client v2 interface. Some methods are unavailable to use such as fetching trends and location, environment fetching methods, and features such as events. Some methods can be recover with OAuth 2 authorization code flow with PKCE with the correct scopes or permissions. Like users.read scope for reading users info which some methods provide a way like :meth:`Client.fetch_user`.
    sleep_after_ratelimit: :class:`bool`
        Indicates to sleep when your client is ratelimited, If set to True it won't raise :class:`TooManyRequests` error but it would print a message indicating to sleep, then it sleeps for how many seconds it needs to sleep, after that it continue to restart the request.
    verify_credentials: :class:`bool`
        Indicates to verify the credentials you specified, this includes consumer_key, consumer_secret, access_token, access_token_secret. make sure to specified all of them in your client, you cannot specified only one of them.

    Attributes
    ------------
    webhook: Optional[:class:`Webhook`]
        Returns the client's main webhook, Returns None if not found.
    environment: Optional[:class:`Environment`]
        Returns the client's main Environment, Returns None if not found.
    webhook_url_path: Optional[:class:`str`]
        Returns the webhook url path, Returns None if not found.


    .. versionadded:: 1.0.0
    """

    def __init__(
        self,
        bearer_token: Optional[str],
        *,
        consumer_key: Optional[str] = None,
        consumer_secret: Optional[str] = None,
        access_token: Optional[str] = None,
        access_token_secret: Optional[str] = None,
        stream: Optional[Stream] = None,
        callback_url: Optional[str] = None,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        use_bearer_only: bool = False,
        sleep_after_ratelimit: bool = False,
        verify_credentials: bool = False,
    ) -> None:
        self.http = HTTPClient(
            bearer_token,
            consumer_key=consumer_key,
            consumer_secret=consumer_secret,
            access_token=access_token,
            access_token_secret=access_token_secret,
            stream=stream,
            callback_url=callback_url,
            client_id=client_id,
            client_secret=client_secret,
            use_bearer_only=use_bearer_only,
            sleep_after_ratelimit=sleep_after_ratelimit,
        )
        self._account_user: Optional[User] = None  # set in account property.
        self.webhook: Optional[Webhook] = None
        self.environment: Optional[Environment] = None
        self.webhook_url_path: Optional[str] = None
        self.executor = self.http.thread_manager.create_new_executor(thread_name="main-executor")
        self.verify_credentials = verify_credentials
        if self.verify_credentials:
            self.http.oauth_session.verify_credentials(raise_error=True)

    def __repr__(self) -> str:
        return "Client({0.account!r})".format(self)

    def account(self, *, update: bool = False) -> Optional[ClientAccount]:
        """Returns :class:`ClientAccount` object which hold the client's informations as a twitter user.

        Parameters
        ------------
        update: :class:`bool`
            Indicates to update the client's account information, setting update to True would make a request to the api and returns a new and updated data everytime. If sets to False, it will either make a request (if used the first time) or use the previous data stored in an instance variable.

            .. versionadded:: 1.5.0


        Returns
        ---------
        Optional[:class:`ClientAccount`]
            This method returns a :class:`ClientAccount` object.


        .. versionadded:: 1.2.0
        """
        account_user = self._account_user
        if account_user is None or update:
            self._set_account_user()
            return self._account_user  # type: ignore
            # The account_user does not change when the function is called. That is why we are returning this.
        return account_user

    def me(self, *, update: bool = False) -> Optional[ClientAccount]:
        """An alias to :meth:`Client.account`.

        Parameters
        ------------
        update: :class:`bool`
            Indicates to update the client's account information, setting update to True would make a request to the api and returns a new and updated data everytime. If sets to False, it will either make a request (if used the first time) or use the previous data stored in an instance variable.

            .. versionadded:: 1.5.0


        Returns
        ---------
        Optional[:class:`ClientAccount`]
            This method returns a :class:`ClientAccount` object.


        .. versionadded:: 1.2.0
        """
        return self.account(update=update)

    def _set_account_user(self) -> None:
        if not self.http.access_token:
            return None

        data = self.http.fetch_me()._User__original_payload
        self._account_user = ClientAccount(data, http_client=self.http)

    def event(self, func: Callable) -> None:
        """A decorator for making an event, the event will be register in the client's internal cache.

        See the :ref:`Event Reference <twitter-api-events>` for the currently documented events.

        Parameters
        ------------
        func: :class:`typing.Callable`
            The function that execute when the event is trigger. The event must be a synchronous function. You must also put the right event name in the function's name, See event reference for full events name.


        Raises
        --------
        TypeError
            The function passed is a coroutine function.


        .. versionadded:: 1.3.5
        """
        if iscoroutinefunction(func):
            raise TypeError("Function passed in event() must NOT be a coroutine function.")
        self.http.events[func.__name__[3:]] = func

    def fetch_user(self, user_id: ID) -> Optional[User]:
        """Fetches a twitter user.

        .. warning::
            This method uses API call and might cause ratelimits if used often! There is always an alternative like :meth:`Client.get_user` from the client's internal cache.

        Parameters
        ------------
        user_id: :class:`ID`
            Represents the user ID that you wish to get info with. If you don't have it you may use `fetch_user_by_username` because it only requires the user's username.

        Returns
        ---------
        :class:`User`
            This method returns a :class:`User` object.


        .. versionadded:: 1.0.0
        """
        return self.http.fetch_user(user_id)

    def fetch_user_by_username(self, username: str) -> Optional[User]:
        """Fetches a twitter user by the user's username.

        .. warning::
            This method uses API call and might cause ratelimits if used often! There is always an alternative like :meth:`Client.get_user` from the client's internal cache.

        Parameters
        ------------
        username: :class:`str`
            Represents the user's username that you wish to get info. A Username usually starts with '@' before any letters. If a username named @Jack, then the username argument must be 'Jack'.

        Returns
        ---------
        :class:`User`
            This method returns a :class:`User` object.


        .. versionadded:: 1.0.0
        """
        return self.http.fetch_user_by_username(username)

    def fetch_tweet(self, tweet_id: ID, *, organic_metrics: bool = False, promoted_metrics: bool = False) -> Tweet:
        """Fetches a tweet.

        .. warning::
            This method uses API call and might cause ratelimits if used often! There is always an alternative like :meth:`Client.get_tweet` from the client's internal cache.

        Parameters
        ------------
        tweet_id: :class:`ID`
            Represents the tweet ID that you wish to get info with.
        organic_metrics: :class:`bool`
            Whether to return an organic metrics data. Grouping of public and non-public metrics attributed to an organic context (posted and viewed in a regular manner). Requires OAuth 1.0a User Context authentication.

            versionadded:: 1.5.0

        promoted_metrics: :class:`bool`
            Whether to return a promoted metrics data. Grouping of public and non-public metrics attributed to a promoted context (posted or viewed as part of an Ads campaign).

            Requires OAuth 1.0a User Context authentication and that the Tweet was promoted in an Ad. Promoted metrics are not included in these counts when a Twitter user is using their own Ads account to promote another Twitter user's Tweets.

            Promoted metrics are included in these counts when:
            * a Twitter user promotes their own Tweets.
            * in an Ads account for a specific handle, the admin for that account may add another Twitter user as an account user so this second account user can promote Tweets for the handle

            versionadded:: 1.5.0

        Raises
        --------
        :class:`ResourceNotFound`
            The tweet was not found

        :class:`DisallowedResource:`
            Examples of why this error raises are:
            * If the tweet is older than 30 days and you fetch it with `organic_metrics` parameter.
            * When you set promoted_metrics to True but the tweet has not been promoted.


        Returns
        ---------
        :class:`Tweet`
            This method returns a :class:`Tweet` object.


        .. versionadded:: 1.0.0
        """
        return self.http.fetch_tweet(tweet_id, organic_metrics=organic_metrics, promoted_metrics=promoted_metrics)

    def fetch_direct_message(self, event_id: ID) -> DirectMessage:
        """Fetches a direct message.

        .. warning::
            This method uses API call and might cause ratelimits if used often! There is always an alternative like :meth:`Client.fetch_direct_message` from the client's internal cache.

        Parameters
        ------------
        event_id: :class:`ID`
            Represents the event's ID that you wish to get info with.

        Returns
        ---------
        :class:`DirectMessage`
            This method returns a :class:`DirectMessage` object.


        .. versionadded:: 1.2.0
        """
        return self.http.fetch_direct_message(event_id)

    def fetch_welcome_message(self, welcome_message_id: ID) -> WelcomeMessage:
        """Fetches a welcome message.

        Parameters
        ------------
        welcome_message_id: :class:`ID`
            Represents the welcome message ID that you wish to get info with.

        Returns
        ---------
        :class:`WelcomeMessage`
            This method returns :class:`WelcomeMessage` object.


        .. versionadded:: 1.3.5
        """
        return self.http.fetch_welcome_message(welcome_message_id)

    def fetch_welcome_message_rule(self, welcome_message_rule_id: ID) -> WelcomeMessageRule:
        """Fetches a welcome message rule.

        Parameters
        ------------
        welcome_message_rule_id: :class:`ID`
            Represents the welcome message rule ID that you wish to get info with.

        Returns
        ---------
        :class:`WelcomeMessageRule`
            This method returns :class:`WelcomeMessageRule` object.


        .. versionadded:: 1.3.5
        """
        return self.http.fetch_welcome_message_rule(welcome_message_rule_id)

    def fetch_space(self, space_id: ID, *, space_host: bool = False) -> Space:
        """Fetches a space.

        .. warning

            This will cause error if `space_host` paramater is True and the client is not the host of the Space!


        Parameters
        ------------
        space_id: :class:`ID`
            Represents the space ID that you wish to get info with.
        space_host: :class:`bool`
            Indicates if the client is the host of the requested space. This is very useful to returns a :class:`Space` object with the 'subscriber_count' data, if sets to False the 'subscriber_count' will returns None. Default to False.

            .. versionchanged:: 1.5.0


        Returns
        ---------
        :class:`Space`
            This method returns a :class:`Space` object.


        .. versionadded:: 1.3.5
        """
        return self.http.fetch_space(space_id, space_host=space_host)

    def fetch_spaces_by_title(
        self, title: str, state: SpaceState = SpaceState.live, *, space_host: bool = False
    ) -> Optional[List[Space]]:
        """Fetches spaces using its title.

        Parameters
        ------------
        title: :class:`ID`
            The space title that you are going use for fetching the space.
        state: :class:`SpaceState`
            The type of state the space has. There are only 2 types: SpaceState.live indicates that the space is live and SpaceState.scheduled indicates the space is not live and scheduled by the host. Default to SpaceState.live,
        space_host: :class:`bool`
            Indicates if the client is the host of the requested space. This is very useful to returns a space with the 'subscriber_count' data, if sets to False the 'subscriber_count' will returns None. Default to False.

            .. versionchanged:: 1.5.0


        Returns
        ---------
        Optional[List[:class:`Space`]]
            This method returns a list of :class:`Space` objects.


        .. versionadded:: 1.3.5
        """
        if state == SpaceState.live or state == SpaceState.scheduled:
            return self.http.fetch_spaces_bytitle(title, state, space_host=space_host)
        else:
            raise UnKnownSpaceState(given_state=state)

    def fetch_list(self, id: ID) -> TwitterList:
        """Fetches a twitter list using its id

        Returns
        ---------
        :class:`List`
            This method returns a :class:`List` object.


        .. versionadded:: 1.5.0
        """
        return self.http.fetch_list(id)

    def fetch_all_environments(self) -> Optional[List[Environment]]:
        """Fetches all the client's environments.

        Returns
        ---------
        Optional[List[:class:`Environment`]]
            Returns a list of :class:`Environment` objects.


        .. versionadded:: 1.5.0
        """
        res = self.http.request("GET", "1.1", "/account_activity/all/webhooks.json")

        return [Environment(data, client=self) for data in res.get("environments")]

    def fetch_message_history(self) -> Optional[MessagePagination]:
        """Returns all Direct Messages (both sent and received) within the last 30 days. Sorted in chronological order.

        Returns
        ---------
        Optional[:class:`MessagePagination`]
            This method returns a :class:`MessagePagination` object.


        .. versionadded:: 1.5.0
        """
        params = {"count": 50}

        res = self.http.request("GET", "1.1", "/direct_messages/events/list.json", auth=True, params=params)

        if not res or not res.get("events"):
            return []

        res = self.http.payload_parser.parse_message_to_pagination_data(res)
        return MessagePagination(
            res, endpoint_request=f"/direct_messages/events/list.jsons", http_client=self.http, params=params
        )

    def fetch_job(self, id: ID) -> Optional[Job]:
        """Fetches a job.

        Parameters
        ------------
        user_id: :class:`ID`
            Represents the job's ID that you wish to get info with.


        .. versionadded:: 1.5.0
        """
        return self.http.fetch_job(id)

    def fetch_jobs(self, type: JobType = JobType.users, status: Optional[JobStatus] = None) -> Optional[List[Job]]:
        """Fetches jobs and minimize the results with the correct :class:`JobType` or :class:`JobStatus`.

        Parameters
        ------------
        type: :class:`JobType`
            The job's type, this parameter minimize the results to only return jobs with the type you specified. Default to :meth:`JobType.users`.
        status: :class:`Jobstatus`
            The job's status, this parameter minimize the results to only return jobs with the status you specified. Default to None.


        .. versionadded:: 1.5.0
        """
        return self.http.fetch_jobs(type, status)

    def tweet(
        self,
        text: str = None,
        *,
        file: Optional[File] = None,
        files: Optional[List[File]] = None,
        poll: Optional[Poll] = None,
        geo: Optional[Union[Geo, str]] = None,
        direct_message_deep_link: Optional[str] = None,
        reply_setting: Optional[Union[ReplySetting, str]] = None,
        quote_tweet: Optional[Union[Tweet, ID]] = None,
        reply_tweet: Optional[Union[Tweet, ID]] = None,
        exclude_reply_users: Optional[List[User, ID]] = None,
        media_tagged_users: Optional[List[User, ID]] = None,
        super_followers_only: bool = False,
    ) -> Optional[Tweet]:
        """Posts a tweet directly to twitter from the given parameters.

        Parameters
        ------------
        text: :class:`str`
            The tweet's text, it will show up as the main text in a tweet.
        file: Optional[:class:`File`]
            Represents a single file attachment. It could be an image, gif, or video. Must be an instance of pytwot.File
        files: Optional[List[:class:`File`]]
            Represents multiple file attachments in a list. It could be an image, gif, or video. the item in the list must also be an instance of pytwot.File
        poll: Optional[:class:`Poll`]
            The poll attachment.
        geo: Optional[Union[:class:`Geo`, :class:`str`]]
            The geo attachment, you can put an object that is an instance of :class:`Geo` or the place ID in a string.
        direct_message_deep_link: Optional[:class:`str`]
            The direct message deep link, It will showup as a CTA(call-to-action) with button attachment.
        reply_setting: Optional[Union[:class:`ReplySetting`, :class:`str`]]
            The reply setting that you can set to minimize users that can reply. If None is specified, the default is set to 'everyone' can reply.
        quote_tweet: Optional[:class:`ID`]
            The tweet or tweet ID you want to quote.
        reply_tweet: Optional[:class:`Tweet`, :class:`ID`]
            The tweet or tweet ID you want to reply. If you have an instance of :class:`Tweet`, you can use the :meth:`Tweet.reply` method rather then using this method.
        exclude_reply_users: Optional[List[:class:`User`, :class:`ID`]]
            A list of users or user ids to be excluded from the reply :class:`Tweet` thus removing a user from a thread, if you dont want to mention a reply with 3 mentions, You can use this argument and provide the user id you don't want to mention.
        media_tagged_users: Optional[List[:class:`User`, :class:`ID`]]
            A list of users or user ids being tagged in the Tweet with Media. If the user you're tagging doesn't have photo-tagging enabled, their names won't show up in the list of tagged users even though the Tweet is successfully created.
        super_followers_only: :class:`bool`
            Allows you to tweet exclusively for super followers.

        Returns
        ---------
        :class:`Tweet`
            This method returns a :class:`Tweet` object.


        .. versionadded:: 1.1.0


        .. versionchanged:: 1.5.0

            Added `files` and `media_tagged_users` arguments.
        """
        return self.http.post_tweet(
            text,
            file=file,
            files=files,
            poll=poll,
            geo=geo,
            quote_tweet=quote_tweet,
            direct_message_deep_link=direct_message_deep_link,
            reply_setting=reply_setting,
            reply_tweet=reply_tweet,
            exclude_reply_users=exclude_reply_users,
            media_tagged_users=media_tagged_users,
            super_followers_only=super_followers_only,
        )

    def create_welcome_message(
        self,
        *,
        name: Optional[str] = None,
        text: Optional[str] = None,
        file: Optional[File] = None,
        quick_reply: Optional[QuickReply] = None,
        cta: Optional[CTA] = None,
    ) -> Optional[WelcomeMessage]:
        """Create a welcome message which you can set with :meth:`WelcomeMessage.set_rule()`.

        Parameters
        ------------
        name: Optional[:class:`str`]
            A human readable name for the Welcome Message
        text: Optional[:class:`str`]
            The welcome message's text. Please do not make this empty if you don't want the text to be blank.
        file: Optional[:class:`File`]:
            Represents a single file attachment. It could be an image, gif, or video. It also have to be an instance of pytwot.File
        quick_reply: Optional[:class:`QuickReply`]
            The message's :class:`QuickReply` attachments.
        cta: Optional[:class:`CTA`]
            The message's :class:`CTA` attachment.

        Returns
        ---------
        :class:`WelcomeMessage`
            This method returns :class:`WelcomeMessage` object.


        .. versionadded:: 1.3.5
        """
        return self.http.create_welcome_message(name=name, text=text, file=file, quick_reply=quick_reply, cta=cta)

    def create_list(self, name: str, *, description: str = "", private: bool = False) -> Optional[TwitterList]:
        """Creates a new list.

        Parameters
        ------------
        name: :class:`str`
            The name of the List you wish to create.
        description: :class:`str`
            Description of the List.
        private: :class:`bool`
            Determine whether the List should be private, default to False.

        Returns
        ---------
        Optional[:class:`List`]
            This method returns a :class:`List` object.


        .. versionadded:: 1.5.0
        """
        twitter_list = self.http.create_list(name, description=description, private=private)
        return self.http.fetch_list(twitter_list.id)

    def create_job(self, type: JobType, *, name: Optional[str] = None, resumable: bool = False) -> Optional[Job]:
        """Creates a Job Compliance.

        Parameters
        ------------
        type: :class:`JobType`
            The type of the job. There's only 2 types of job, :meth:`JobType.tweets' and :class:`JobType.users`. You can upload tweet ids if use :meth:`JobType.tweets` otherwise you can use user ids. You cannot mix tweet ids with user ids!
        name: :class:`str`
            The human readable name for the Job, default to None.
        resumable: :class:`bool`
                Specifies whether to enable the upload URL with support for resumable uploads. If true, this endpoint will return a pre-signed URL with resumable uploads enabled. Default to False

        Returns
        ---------
        Optional[:class:`Job`]
            This method returns :class:`Job` object


        .. versionadded:: 1.5.0
        """
        return self.http.create_job(type, name=name, resumable=resumable)

    def create_custom_profile(self, name: str, file: File) -> Optional[CustomProfile]:
        """Create a custom profile

        Parameters
        ------------
        name: :class:`str`
            The author's custom name.
        file: :class:`File`
            The media file that's associate with the profile.

        Returns
        ---------
        :class:`CustomProfile`
            This method returns a :class:`CustomProfile` object.
        """
        return self.http.create_custom_profile(name, file)

    def search_geos(
        self,
        query: str,
        max_result: Optional[ID] = None,
        *,
        lat: Optional[int] = None,
        long: Optional[int] = None,
        ip: Optional[ID] = None,
        granularity: Granularity = Granularity.neighborhood,
    ) -> Geo:
        """Search geo-locations with the given arguments.

        Parameters
        ------------
        query: :class:`str`
            Free-form text to match against while executing a geo-based query, best suited for finding nearby locations by name. Remember to URL encode the query.
        max_results: Optional[:class:`ID`]
            A hint as to the number of results to return. This does not guarantee that the number of results returned will equal max_results, but instead informs how many "nearby" results to return. Ideally, only pass in the number of places you intend to display to the user here.
        lat: :class:`int`
            The latitude to search around. This parameter will be ignored unless it is inside the range -90.0 to +90.0 (North is positive) inclusive. It will also be ignored if there isn't a corresponding long parameter.
        long: :class:`int`
            The longitude to search around. The valid ranges for longitude are -180.0 to +180.0 (East is positive) inclusive. This parameter will be ignored if outside that range, if it is not a number, if geo_enabled is turned off, or if there is not a corresponding lat parameter.
        ip: :class:`ID`
            An IP address. Used when attempting to fix geolocation based off of the user's IP address.
        granularity: :class:`str`
            This is the minimal granularity of place types to return and must be one of: neighborhood, city, admin or country. If no granularity is provided for the request neighborhood is assumed. Setting this to city, for example, will find places which have a type of city, admin or country.

        Returns
        ---------
        List[:class:`Geo`]
            This method return a list of :class:`Geo` objects.


        .. versionadded:: 1.3.5
        """
        return self.http.search_geo(query, max_result, lat=lat, long=long, ip=ip, granularity=granularity)

    def search_trend_with_place(self, woeid: ID, exclude: Optional[str] = None) -> Optional[List[Trend]]:
        """Search trends with woeid.

        .. note::
            You can find woeid information through :meth:`Location.woeid` with :meth:`Client.search_trend_locations` or :meth:`Client.search_trend_closest`.

        Parameters
        ------------
        woeid: :class:`ID`
            "where on earth identifier" or WOEID, which is a legacy identifier created by Yahoo and has been deprecated. Twitter API v1.1 still uses the numeric value to identify town and country trend locations. Example WOEID locations include: Worldwide: 1 UK: 23424975 Brazil: 23424768 Germany: 23424829 Mexico: 23424900 Canada: 23424775 United States: 23424977 New York: 2459115.
        exclude: Optional[:class:`str`]
            Setting this equal to hashtags will remove all hashtags from the trends list.

        Returns
        ---------
        Optional[List[:class:`Trend`]]
            This method returns a list of :class:`Trend` objects.


        .. versionadded:: 1.5.0
        """
        res = self.http.request(
            "GET",
            "1.1",
            "/trends/place.json",
            params={"id": str(woeid), "exclude": exclude},
            auth=True,
        )

        return [Trend(**data) for data in res[0].get("trends")]

    def search_trend_locations(self) -> Optional[List[Location]]:
        """Search locations that Twitter has trending topic information.

        Returns
        ---------
        Optional[List[:class:`Location`]]
            This method returns a list of :class:`Location` objects.


        .. versionadded:: 1.5.0
        """
        res = self.http.request("GET", "1.1", "/trends/available.json", auth=True)
        for index, data in enumerate(res):
            res[index] = self.http.payload_parser.parse_trend_location_payload(data)

        return [Location(**data) for data in res]

    def search_trend_closest(self, lat: int, long: int) -> Optional[List[Location]]:
        """Search the rend closest to the lat and long.

        parameters
        ------------
        lat: :class:`int`
            If provided with a long parameter the available trend locations will be sorted by distance, nearest to furthest, to the coordinate pair. The valid ranges for longitude is -180.0 to +180.0 (West is negative, East is positive) inclusive.
        long: :class:`int`
            If provided with a lat parameter the available trend locations will be sorted by distance, nearest to furthest, to the coordinate pair. The valid ranges for longitude is -180.0 to +180.0 (West is negative, East is positive) inclusive -122.400612831116

        Returns
        ---------
        Optional[List[:class:`Location`]]
            This method returns a list of :class:`Location` objects.


        .. versionadded:: 1.5.0
        """
        res = self.http.request(
            "GET",
            "1.1",
            "/trends/available.json",
            params={"lat": lat, "long": long},
            auth=True,
        )

        return [Location(**data) for data in res]

    def search_recent_tweet(
        self,
        query: str,
        *,
        max_results: int = 10,
        start_time: Optional[datetime.datetime] = None,
        end_time: Optional[datetime.datetime] = None,
        since_id: Optional[ID] = None,
        until_id: Optional[ID] = None,
        sort_by_relevancy: bool = False,
    ) -> List[Tweet]:
        """Searches tweet from the last seven days that match a search query.

        Parameters
        ------------
        query: :class:`str`
            One query for matching Tweets.
        max_results: :class:`int`
            The maximum number of search results to be returned by a request. A number between 10 and 100. By default, the method will returns 10 results.
        start_time: Optional[:class:`datetime.datetime`]
            This will make sure the tweets created datetime is after that specific time.
        end_time: Optional[:class:`datetime.datetime`]
            This will make sure the tweets created datetime is before that specific time.
        since_id: Optional[`ID`]
            Returns results with a Tweet ID greater than (that is, more recent than) the specified 'since' Tweet ID. Only the 3200 most recent Tweets are available. The result will exclude the since_id. If the limit of Tweets has occurred since the since_id, the since_id will be forced to the oldest ID available.
        until_id: Optional[`ID`]
            Returns results with a Tweet ID less less than (that is, older than) the specified 'until' Tweet ID. Only the 3200 most recent Tweets are available. The result will exclude the until_id. If the limit of Tweets has occurred since the until_id, the until_id will be forced to the most recent ID available.
        sort_by_relevancy: :class:`bool`
            This parameter is used to specify the order in which you want the Tweets returned. If sets to True, tweets will be order by relevancy, else it sets to recency. Default to False.

        Returns
        ---------
        Union[:class:`TweetPagination`, :class:`list`]
            This method returns a list of :class:`Tweet` objects.


        .. versionadded:: 1.5.0
        """
        if (
            not isinstance(start_time, datetime.datetime)
            and start_time
            or not isinstance(end_time, datetime.datetime)
            and end_time
        ):
            raise ValueError("start_time or end_time must be a datetime object!")

        params = {
            "expansions": TWEET_EXPANSION,
            "user.fields": USER_FIELD,
            "media.fields": MEDIA_FIELD,
            "place.fields": PLACE_FIELD,
            "poll.fields": POLL_FIELD,
            "tweet.fields": TWEET_FIELD,
            "query": query,
            "max_results": max_results,
        }

        if start_time:
            params["start_time"] = start_time.isoformat()
        if end_time:
            params["end_time"] = end_time.isoformat()
        if since_id:
            params["since_id"] = str(since_id)
        if until_id:
            params["until_id"] = str(until_id)
        if sort_by_relevancy:
            params["sort_order"] = "relevancy"

        res = self.http.request("GET", "2", "/tweets/search/recent", params=params)

        return [Tweet(data, http_client=self.http) for data in res.get("data")]

    def search_all_tweet(
        self,
        query: str,
        *,
        max_results: int = 10,
        start_time: Optional[datetime.datetime] = None,
        end_time: Optional[datetime.datetime] = None,
        since_id: Optional[ID] = None,
        until_id: Optional[ID] = None,
        sort_by_relevancy: bool = False,
    ) -> List[Tweet]:
        """Searches all tweet from the complete history of public Tweets matching a search query; since the first Tweet was created March 26, 2006. Only available to those users who have been approved for Academic Research access.

        Parameters
        ------------
        query: :class:`str`
            One query for matching Tweets.
        max_results: :class:`int`
            The maximum number of search results to be returned by a request. A number between 10 and 100. By default, the method will returns 10 results.
        start_time: Optional[:class:`datetime.datetime`]
            This will make sure the tweets created datetime is after that specific time.
        end_time: Optional[:class:`datetime.datetime`]
            This will make sure the tweets created datetime is before that specific time.
        since_id: Optional[`ID`]
            Returns results with a Tweet ID greater than (that is, more recent than) the specified 'since' Tweet ID. Only the 3200 most recent Tweets are available. The result will exclude the since_id. If the limit of Tweets has occurred since the since_id, the since_id will be forced to the oldest ID available.
        until_id: Optional[`ID`]
            Returns results with a Tweet ID less less than (that is, older than) the specified 'until' Tweet ID. Only the 3200 most recent Tweets are available. The result will exclude the until_id. If the limit of Tweets has occurred since the until_id, the until_id will be forced to the most recent ID available.
        sort_by_relevancy: :class:`bool`
            This parameter is used to specify the order in which you want the Tweets returned. If sets to True, tweets will be order by relevancy, else it sets to recency. Default to False.

        Returns
        ---------
        Union[:class:`TweetPagination`, :class:`list`]
            This method returns a list of :class:`Tweet` objects.


        .. versionadded:: 1.5.0
        """
        if (
            not isinstance(start_time, datetime.datetime)
            and start_time
            or not isinstance(end_time, datetime.datetime)
            and end_time
        ):
            raise ValueError("start_time or end_time must be a datetime object!")

        params = {
            "expansions": TWEET_EXPANSION,
            "user.fields": USER_FIELD,
            "media.fields": MEDIA_FIELD,
            "place.fields": PLACE_FIELD,
            "poll.fields": POLL_FIELD,
            "tweet.fields": TWEET_FIELD,
            "query": query,
            "max_results": max_results,
        }

        if start_time:
            params["start_time"] = start_time.isoformat()
        if end_time:
            params["end_time"] = end_time.isoformat()
        if since_id:
            params["since_id"] = str(since_id)
        if until_id:
            params["until_id"] = str(until_id)
        if sort_by_relevancy:
            params["sort_order"] = "relevancy"

        res = self.http.request("GET", "2", "/tweets/search/all", params=params)

        return [Tweet(data, http_client=self.http) for data in res.get("data")]

    def get_user(self, user_id: ID) -> Optional[User]:
        """Gets a user through the client internal user cache. Return None if the user is not in the cache.

        .. note::
            Users will get cache with several conditions:
                * Users return from a method such as :meth:`Client.fetch_user`.
                * The client interacts with other users such as sending a message to another user through :meth:`User.send` and many more
                * The subscription users interact with other users such as sending message from the subscription user to another user (This condition only applies if you use :meth:`Client.listen` at the very end of the file)

        Parameters
        ------------
        user_id: :class:`ID`
            The ID of a user that you want to get.

        Raises
        --------
        ValueError:
            Raised when the user_id argument is not an integer or a string of digits.

        Returns
        ---------
        :class:`User`
            This method returns a :class:`User` object or None if the user was not found.


        .. versionadded:: 1.5.0
        """
        try:
            user_id = int(user_id)
        except ValueError:
            raise ValueError("user_id must be an integer or a string of digits.")

        return self.http.user_cache.get(user_id)

    def get_tweet(self, tweet_id: ID) -> Optional[Tweet]:
        """Gets a tweet through the client internal tweet cache. Return None if the tweet is not in the cache.

        .. note::
            Tweets will get cache with several conditions:
                * Tweets send by the client.
                * Tweets send by the subscription users (This condition only applies if you use :meth:`Client.listen` at the very end of the file).
                * Tweets return from a method such as: :meth:`Client.fetch_tweet`

        Parameters
        ------------
        tweet_id: :class:`ID`
            The ID of a tweet that you want to get.

        Raises
        --------
        ValueError:
            Raised when the tweet_id argument is not an integer or a string of digits.

        Returns
        ---------
        :class:`Tweet`
            This method returns a :class:`Tweet` object or None if the tweet was not found.


        .. versionadded:: 1.2.0
        """
        try:
            tweet_id = int(tweet_id)
        except ValueError:
            raise ValueError("tweet_id must be an integer or a string of digits.")

        return self.http.tweet_cache.get(tweet_id)

    def get_direct_message(self, event_id: ID) -> Optional[DirectMessage]:
        """Get a direct message through the client message cache. Returns None if the message is not in the cache.

        .. note::
            Messages will get cache with several conditions:
                * Messages send by the client.
                * Messages return from a method such as: :meth:`Client.fetch_direct_message`
                * The subscription users interact with other users such as sending message from the subscription user to another user (This condition only applies if you use :meth:`Client.listen` at the very end of the file).

        Parameters
        ------------
        event_id: :class:`ID`
            The event ID of the Direct Message event that you want to get.

        Returns
        ---------
        :class:`DirectMessage`
            This method returns a :class:`DirectMessage` object.


        .. versionadded:: 1.2.0
        """
        try:
            event_id = int(event_id)
        except ValueError:
            raise ValueError("Event id must be an integer or a :class:`str`ing of digits.")

        return self.http.message_cache.get(event_id)

    def stream(self, *, dry_run: bool = False) -> None:
        """Stream realtime in twitter for tweets! This method use the stream argument in :meth:`request.get` for streaming in one of the stream endpoint that twitter api provides. If you want to use this method, make sure to provides the stream kwarg in your :class:`Client` instance and make an on_stream event to get the stream's tweet data and connection,
        example:

        .. code-block:: py

            import pytwot

            stream = pytwot.Stream()
            stream.add_rule("pytwot") #this make sure to only return tweets that has pytwot keyword in it.

            client = pytwot.Client(
                ...
                stream=stream
            )

            @client.event
            def on_stream(tweet, connection):
                ... #Do what you want with tweet and stream connection you got.

            client.stream()

        You can also add rules and specified which tweets must be retrieve via the tweet's characteristic features.

        Parameters
        ------------
        dry_run: :class:`bool`
            Indicates if you want to debug your rule's operator syntax.


        .. versionadded:: 1.3.5
        """
        if not self.http.stream:
            raise TypeError("'stream' argument is missing in client!")

        try:
            self.http.stream.connect(dry_run=dry_run)
        except KeyboardInterrupt:
            print("\nKeyboardInterrupt: Exit stream.")

    def listen(
        self,
        app: Flask,
        *,
        url: str,
        env_label: str,
        sleep_for: Union[int, float] = 0.50,
        ngrok: bool = False,
        **kwargs: Any,
    ):
        """Listen to upcoming account activity events send by twitter to your flask's url. You can use the rest of Flask arguments like port or host via the kwargs argument.

        .. note::
            For the time being, we only support Flask for the app argument! If you want to use your own web application url, consider using :meth:`Client.listen_to`.

        Parameters
        ------------
        app: :class:`flask.Flask`
            Your flask application.
        url: :class:`str`
            a kwarg that the webhook url aka your flask's web application url. This completely up to you, e.g https://your-website.domain/webhook/twitter.
        env_label: :class:`str`
            a kwarg that the environment's label.
        sleep_for: Union[:class:`int`, :class:`float`]
            a kwarg that ensure the flask application is running before triggering a CRC by sleeping after starting a thread. Default to 0.50.
        ngrok: :class:`bool`
            a kwarg that indicates to use ngrok for tunneling your localhost. This usually uses for users that use localhost url.
        disabled_log: :class:`bool`
            A kwarg that indicates to disable flask's log so it does not print the request process in your terminal, this also will disable `werkzeug` log.
        make_new: :class:`bool`
            A kwarg indicates to make a new webhook url when the api can't find the url passed. Default to True.


        .. versionadded:: 1.5.0
        """
        if not isinstance(app, Flask):
            raise pytwotException("App argument must be an instance of flask.Flask!")

        disabled_log: bool = kwargs.pop("disabled_log", False)
        make_new: bool = kwargs.pop("make_new", True)
        environments = self.fetch_all_environments()

        if disabled_log:
            app.logger.disabled = True
            log = logging.getLogger("werkzeug")
            log.disabled = True

        for env in environments:
            if env.label == env_label:
                self.environment = env

            for webhook in env.webhooks:
                if url == webhook.url:
                    self.webhook_url_path = urlparse(webhook.url).path
                    self.webhook = webhook
                    self.environment = env
                    break

        try:
            thread = threading.Thread(
                target=app.run,
                name="client-listen-method:thread_session=LISTEN-SESSION",
                kwargs=kwargs,
            )

            if not self.webhook and not ngrok:
                self.webhook_url_path = urlparse(url).path

            elif not self.webhook and ngrok:
                ...  # TODO add ngrok support

            @app.route(self.webhook_url_path, methods=["POST", "GET"])
            def webhook_receive():
                if request.method == "GET":
                    _log.info("Attempting to respond a CRC.")
                    crc = request.args["crc_token"]

                    validation = hmac.new(
                        key=bytes(self.http.consumer_secret, "UTF-8"),
                        msg=bytes(crc, "UTF-8"),
                        digestmod=hashlib.sha256,
                    )
                    digested = base64.b64encode(validation.digest())

                    response = {"response_token": "sha256=" + format(str(digested)[2:-1])}

                    return json.dumps(response)

                json_data = request.get_json()
                _log.debug(f"An event triggered! {json_data}")
                self.executor.submit(self.http.handle_events, payload=json_data)
                return ("", HTTPStatus.OK)

            check = not self.webhook and self.webhook_url_path
            if check and make_new:
                thread.start()
                time.sleep(sleep_for)
                webhook = self.environment.register_webhook(
                    url
                )  # Register a new webhook url if no webhook found also if make_new is True.
                self.webhook = webhook
                self.environment = env
                self.environment.add_my_subscription()
                ids = self.environment.fetch_all_subscriptions()
                users = self.http.fetch_users(ids)
                for user in users:
                    self.http.user_cache[user.id] = user

                _log.debug(
                    f"Listening for events! user cache filled at {len(self.http.user_cache)} users! flask application is running with url: {url}({self.webhook_url_path}).\n Ngrok: {ngrok}\nMake a new webhook when not found: {make_new}\n In Environment: {repr(self.environment)} with webhook: {repr(self.webhook)}."
                )

            elif check and not make_new:
                raise pytwotException(
                    f"Cannot find url passed: {url} Invalid url passed, please check the spelling of your url"
                )

            else:
                thread.start()
                time.sleep(sleep_for)
                self.webhook.trigger_crc()
                ids = self.environment.fetch_all_subscriptions()
                users = self.http.fetch_users(ids)
                for user in users:
                    self.http.user_cache[user.id] = user

                _log.debug(
                    f"Listening for events! user cache filled at {len(self.http.user_cache)} users! flask application is running with url: {url}({self.webhook_url_path}).\n Ngrok: {ngrok}\nMake a new webhook when not found: {make_new}\n In Environment: {repr(self.environment)} with webhook: {repr(self.webhook)}."
                )
            thread.join()
        except Exception as e:
            _log.warn(f"An error was caught during listening: {e}")
            raise e

        finally:
            _log.debug(f"Stop listening due to internal/external problem!")

    def listen_to(self, url: str, *, env_label: str, ngrok: bool = False, make_new: bool = True):
        """Listen to upcoming account activity events send by twitter to a web application url. This method differ from :meth:`Client.listen`, this method doesn't use the flask's web application url, rather your web application url. This is good for people that want to implement their web application outside flask.

        .. warning::
            With this method, you have to make your own CRC and event handlers in your web application. For the time being, the documentation doesn't provides information for the handlers, either go to twitter documentation about account activity api or wait until we write the documentation.

        Parameters
        ------------
        url: :class:`str`
            The webhook url. This completely up to you, e.g https://your-website.domain/webhook/twitter.
        env_label: :class:`str`
            The environment's label.
        ngrok: :class:`bool`
            indicates to use ngrok for tunneling your localhost. This is usually use for users that use localhost url.
        make_new: :class:`bool`
            A kwarg indicates to make a new webhook url when the api can't find the url passed. Default to True.


        .. versionadded:: 1.5.0
        """
        environments = self.fetch_all_environments()

        for env in environments:
            if env.label == env_label:
                self.environment = env

            for webhook in env.webhooks:
                if url == webhook.url:
                    self.webhook_url_path = urlparse(webhook.url).path
                    self.webhook = webhook
                    self.environment = env
                    break

        if not self.webhook and not ngrok:
            self.webhook_url_path = urlparse(webhook.url).path

        elif self.webhook and ngrok:
            ...  # TODO add ngrok support

        check = not self.webhook and self.webhook_url_path
        if check and make_new:
            webhook = self.environment.register_webhook(
                url
            )  # Register a new webhook url if no webhook found also if make_new is True.
            self.webhook = webhook
            self.environment = env
            self.environment.add_my_subscription()
            ids = self.environment.fetch_all_subscriptions()
            users = self.http.fetch_users(ids)
            for user in users:
                self.http.user_cache[user.id] = user

            _log.debug(
                f"Listening for events! user cache filled at {len(self.http.user_cache)} users! flask application is running with url: {url}({self.webhook_url_path}).\n Ngrok: {ngrok}\nMake a new webhook when not found: {make_new}\n In Environment: {repr(self.environment)} with webhook: {repr(self.webhook)}."
            )

        elif check and not make_new:
            raise pytwotException(
                f"Cannot find url passed: {url} Invalid url passed, please check the spelling of your url"
            )

        else:
            self.webhook.trigger_crc()
            ids = self.environment.fetch_all_subscriptions()
            users = self.http.fetch_users(ids)
            for user in users:
                self.http.user_cache[user.id] = user

            _log.debug(
                f"Listening for events! user cache filled at {len(self.http.user_cache)} users! flask application is running with url: {url}({self.webhook_url_path}).\n Ngrok: {ngrok}\nMake a new webhook when not found: {make_new}\n In Environment: {repr(self.environment)} with webhook: {repr(self.webhook)}."
            )



"""
The MIT License (MIT)

Copyright (c) 2021-present UnrealFar & TheGenocides
Copyright (c) 2023-present Sengolda

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import datetime
from typing import Union, Optional
from .user import User
from .tweet import Tweet
from .enums import ActionEventType, UserActionEventType
from .type import Payload
from .utils import time_parse_todt

# Events type


class Event:
    """The base class of all event, this provides a type property that the event belongs to.

    .. versionadded:: 1.5.0
    """

    __slots__ = ("_payload", "_type")

    def __init__(self, data: Payload):
        self._type = list(data.keys())[1]
        self._payload = data.get(self._type)[0]

    @property
    def type(self) -> Union[UserActionEventType, ActionEventType]:
        """Union[:class:`UserActionEventType`, :class:`ActionEventType`]: Returns the event's action type.

        .. versionadded:: 1.5.0
        """
        try:
            return ActionEventType(self._type)
        except ValueError:
            return UserActionEventType(self._type)

    @property
    def payload(self) -> Payload:
        """Returns the event payload.

        .. versionadded:: 1.5.0
        """
        return self._payload


class UserActionEvent(Event):
    """Represents a user action event, this could be but not limited to: follow or unfollow. This is also a parent class to all UserAction events, this class inherits :class:`Event`.

    .. versionadded:: 1.5.0
    """

    @property
    def created_at(self) -> datetime.datetime:
        """:class:`datetime.datetime`: Returns a datetime.datetime object with the action's created timestamp.

        .. versionadded:: 1.5.0
        """
        return datetime.datetime.fromtimestamp(int(self.payload.get("created_timestamp")) / 1000)

    @property
    def target(self) -> User:
        """:class:`User`: Returns the user that was targeted by the source.

        .. versionadded:: 1.5.0
        """
        return self.payload.get("target")

    @property
    def source(self) -> User:
        """:class:`User`: Returns a user object that trigger this event.

        .. versionadded:: 1.5.0
        """
        return self.payload.get("source")

    @property
    def author(self):
        """:class:`User`: An alias to source.

        .. versionadded:: 1.5.0
        """
        return self.source


class DirectMessageEvent(Event):
    def __init__(self, data: Payload, *, http_client: object):
        super().__init__(data)
        self.http_client = http_client

    @property
    def created_at(self) -> datetime.datetime:
        """:class:`datetime.datetime`: Returns a datetime.datetime object with the action's created timestamp.

        .. versionadded:: 1.5.0
        """
        return datetime.datetime.fromtimestamp(int(self.payload.get("created_at") / 1000))

    @property
    def recipient(self) -> User:
        """:class:`User`: Returns the user that :meth:`DirectMessageEvent.sender` interacted.

        .. versionadded:: 1.5.0
        """
        return self.payload.get("target", {}).get("recipient")

    @property
    def sender(self) -> User:
        """:class:`User`: Returns the user that trigger the event.

        .. versionadded:: 1.5.0
        """
        return self.payload.get("target", {}).get("sender", None)


# Events


class DirectMessageTypingEvent(DirectMessageEvent):
    """Represents a typing event object for `on_typing` event, this inherits :class:`DirectMessageEvent`. This object contains the event information that twitter posts through the webhook url.

    .. versionadded:: 1.5.0
    """

    @property
    def typer(self) -> User:
        """:class:`User`: An alias to :meth:`DirectMessageEvent.sender`.

        .. versionadded:: 1.5.0
        """
        return self.sender


class DirectMessageReadEvent(DirectMessageEvent):
    """Represents a direct message read event object for `on_direct_message_read` event, this inherits :class:`DirectMessageEvent`. This object contains information that twitter posts through the webhook url.

    .. versionadded:: 1.5.0
    """

    @property
    def reader(self) -> Optional[User]:
        """:class:`User`: An alias to :meth:`DirectMessageEvent.sender`.

        .. versionadded:: 1.5.0
        """
        return self.sender

    @property
    def last_read_message(self) -> Optional[int]:
        """Optional[:class:`int`]: Returns the last message that got read, returns the event id if none found.

        .. versionadded:: 1.5.0
        """
        event_id = self.payload.get("last_read_event_id")
        message = self.http_client.message_cache.get(int(event_id))
        if not message:
            return event_id
        return message


class TweetFavoriteActionEvent(Event):
    """Represents a tweet favorite event object for `on_tweet_favorite` event, this inherits :class:`Event`. This object contains information that twitter posts through the webhook url.

    .. versionadded:: 1.5.0
    """

    @property
    def id(self) -> str:
        """:class:`str`: Returns the action's unique ID.

        .. versionadded:: 1.5.0
        """
        return self.payload.get("id")

    @property
    def created_at(self) -> datetime.datetime:
        """:class:`datetime.datetime`: Returns a datetime.datetime object with the action's created timestamp.

        .. versionadded:: 1.5.0
        """
        return datetime.datetime.fromtimestamp(int(self.payload.get("timestamp_ms")) / 1000)

    @property
    def favorited_tweet(self) -> Tweet:
        """:class:`Tweet`: Returns the favorited tweet.

        .. versionadded:: 1.5.0
        """
        return self.payload.get("tweet")

    @property
    def liker(self) -> User:
        """:class:`User`: Returns the user who favorited the tweet.

        .. versionadded:: 1.5.0
        """
        return self.payload.get("liker")


class UserRevokeEvent(Event):
    """Represents a revoke access event by the subcription user, this inherits :class:`Event`. This object contains information that twitter posts through the webhook url.

    .. versionadded:: 1.5.0
    """

    def __init__(self, data: Payload):
        super().__init__(data.get("revoke"))

    @property
    def revoked_at(self) -> datetime.datetime:
        """:class:`datetime.datetime`: Returns a datetime.datetime object with the action's created timestamp.

        .. versionadded:: 1.5.0
        """
        return time_parse_todt(self.payload.get("date_time"))

    @property
    def app_id(self) -> int:
        """:class:`int`: Returns an application id who the user revoked from.

        .. versionadded:: 1.5.0
        """
        return int(self.payload.get("target").get("app_id"))

    @property
    def user_id(self) -> int:
        """:class:`int`: Returns a user id who revoked the access.

        .. versionadded:: 1.5.0
        """
        return int(self.payload.get("source").get("user_id"))


class UserFollowActionEvent(UserActionEvent):
    """Represents a follow action event by user or of user, This inherits :class:`UserActionEvent`. This object contains information that twitter posts through the webhook url.

    .. versionadded:: 1.5.0
    """

    @property
    def follower(self) -> User:
        """:class:`User`: An alias to :meth:`UserActionEvent.source`. The user who followed/unfollowed the target.

        .. versionadded:: 1.5.0
        """
        return self.source


class UserBlockActionEvent(UserActionEvent):
    """Represents a block action event by user, This inherits :class:`UserActionEvent`. This object contains information that twitter posts through the webhook url.

    .. versionadded:: 1.5.0
    """

    @property
    def blocker(self) -> User:
        """:class:`User`: An alias to :meth:`UserActionEvent.source`. The user who blocked/unblocked the target.

        .. versionadded:: 1.5.0
        """
        return self.source


class UserMuteActionEvent(UserActionEvent):
    """Represents a mute action event by user, This inherits :class:`UserActionEvent`. This object contains information that twitter posts through the webhook url.

    .. versionadded:: 1.5.0
    """

    @property
    def muter(self) -> User:
        """:class:`User`: An alias to :meth:`UserActionEvent.source`. The user who muted/unmuted the target.

        .. versionadded:: 1.5.0
        """
        return self.source


class UserUnfollowActionEvent(UserFollowActionEvent):
    """Represents an unfollow action event by user, This inherits :class:`UserFollowActionEvent`. This object contains information that twitter posts through the webhook url.

    .. versionadded:: 1.5.0
    """

    ...


class UserUnblockActionEvent(UserBlockActionEvent):
    """Represents an unblock action event by user, This inherits :class:`UserBlockActionEvent`. This object contains information that twitter posts through the webhook url.

    .. versionadded:: 1.5.0
    """

    ...


class UserUnmuteActionEvent(UserMuteActionEvent):
    """Represents an unmute action event by user, This inherits :class:`UserMuteActionEvent`. This object contains information that twitter posts through the webhook url.

    .. versionadded:: 1.5.0
    """

    ...



"""
The MIT License (MIT)

Copyright (c) 2021-present UnrealFar & TheGenocides
Copyright (c) 2023-present Sengolda

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from dataclasses import dataclass
from typing import List, Optional

__all__ = ("Embed", "EmbedImage")


@dataclass
class EmbedImage:
    """Represents a dataclass for an embed image.

    .. versionadded: 1.1.3

    .. versionchanged: 1.5.0

        Made as a dataclass rather then a standalone class.
    """

    url: str
    width: int
    height: int


@dataclass
class Embed:
    """Represents a dataclass for embedded urls in a tweet.

    .. versionadded: 1.1.3


    .. versionchanged: 1.5.0

        Made as a dataclass rather then a standalone class.
    """

    title: Optional[str] = None
    description: Optional[str] = None
    start: Optional[int] = None
    end: Optional[int] = None
    url: Optional[str] = None
    expanded_url: Optional[str] = None
    display_url: Optional[str] = None
    unwound_url: Optional[str] = None
    status_code: Optional[int] = None
    images: Optional[List[EmbedImage]] = None



"""
The MIT License (MIT)

Copyright (c) 2021-present UnrealFar & TheGenocides
Copyright (c) 2023-present Sengolda

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import datetime
import io
import os
from typing import Any, Dict, List, Optional, Union, TYPE_CHECKING

from .dataclass import PollOption, Option, Button
from .enums import ButtonType, MediaType
from .utils import time_parse_todt, guess_mimetype, convert
from .errors import pytwotException
from .constants import LANGUAGES_CODES
from .objects import Comparable
from .dataclass import NonPublicMediaMetrics, OrganicMediaMetrics, PromotedMediaMetrics

if TYPE_CHECKING:
    from .type import ID
    from .http import HTTPClient

__all__ = ("Media", "Poll", "QuickReply", "Geo", "CTA", "File", "SubFile")


class Media:
    """Represents a media attachment in a message.

    Note that this object is different than :class:`File`. While you need :class:`File` to send attachment (this includes .png, .jpeg, .gif and .mp4) to a message, Media is like the final result of that File (after you initialize it through upload endpoints).


    """

    __slots__ = (
        "http_client",
        "_payload",
        "_url",
        "_preview_image_url",
        "_media_key",
        "_type",
        "_width",
        "_height",
        "_public_metrics",
        "_non_public_metrics",
        "_organic_metrics",
        "_promoted_metrics",
    )

    def __init__(self, data: Dict[str, Any], *, http_client: HTTPClient):
        self.http_client = http_client
        self._payload = data
        self._url = self._payload.get("url")
        self._preview_image_url = self._payload.get("preview_image_url")
        self._media_key = self._payload.get("media_key")
        self._type = MediaType(self._payload.get("type"))
        self._width, self._height = self._payload.get("width"), self._payload.get("height")
        self._public_metrics = self._payload.get("public_metrics")
        self._non_public_metrics = self._payload.get("non_public_metrics")
        self._organic_metrics = self._payload.get("organic_metrics")
        self._promoted_metrics = self._payload.get("promoted_metrics")

        if self._non_public_metrics:
            non_public_metrics = self.http_client.payload_parser.parse_metric_data(self._non_public_metrics)
            self._non_public_metrics = NonPublicMediaMetrics(**non_public_metrics)

        if self._organic_metrics:
            organic_metrics = self.http_client.payload_parser.parse_metric_data(self._organic_metrics)
            self._organic_metrics = OrganicMediaMetrics(**organic_metrics)

        if self._promoted_metrics:
            promoted_metrics = self.http_client.payload_parser.parse_metric_data(self._promoted_metrics)
            self._promoted_metrics = PromotedMediaMetrics(**promoted_metrics)

    @property
    def url(self) -> str:
        """:class:`str`: Returns the image's url, this method is only available if the media type is :class:`MediaType.photo`. If the media type is :class:`MediaType.video` consider using :class:`Media.preview_image_url`."""
        return self._url

    @property
    def preview_image_url(self) -> str:
        """:class:`str`: Returns the video's preview image url, This is only available when the media type is a :class:`MediaType.video` which is for video only."""
        return self._preview_image_url

    @property
    def key(self) -> str:
        """:class:`str`: Returns the media's key"""
        return self._media_key

    @property
    def type(self) -> MediaType:
        """:class:`str`: Returns the image's type in a :meth:`MediaType` object."""
        return self._type

    @property
    def width(self) -> Optional[int]:
        """Optional[:class:`int`]: Returns the image's width"""
        return convert(self._width, int)

    @property
    def height(self) -> Optional[int]:
        """Optional[:class:`int`]: Returns the image's height"""
        return convert(self._height, int)

    @property
    def view_count(self) -> Optional[int]:
        """Optional[:class:`int`]: Returns the media's `view_count` if the media count as a `video` type.

        .. versionadded:: 1.5.0
        """
        if not self._public_metrics:
            return None
        return convert(self._public_metrics["view_count"], int)

    @property
    def non_public_metrics(self) -> Optional[NonPublicMediaMetrics]:
        """Optional[:class:`NonPublicMediaMetrics`]: The media's metrics that are not available for anyone to view on Twitter, such as `video view quartiles`.

        .. versionadded:: 1.5.0
        """
        return self._non_public_metrics

    @property
    def organic_metrics(self) -> Optional[OrganicMediaMetrics]:
        """Optional[:class:`OrganicMediaMetrics`]: The media's metrics in organic context (posted and viewed in a regular manner), such as `video view quartiles`.

        .. versionadded:: 1.5.0
        """
        return self._organic_metrics

    @property
    def promoted_metrics(self) -> Optional[PromotedMediaMetrics]:
        """Optional[:class:`PromotedMediaMetrics`]: The tweet's metrics in promoted context (posted or viewed as part of an Ads campaign), such as `video view quartiles`.

        .. versionadded:: 1.5.0
        """
        return self._promoted_metrics


class Poll(Comparable):
    """Represents a Poll attachment in a tweet.

    .. describe:: x == y

        Check if one Poll's id is equal to another.


    .. describe:: x != y

        Check if one Poll's id is not equal to another.


    .. describe:: len(x)

        returns how many options the poll have.

    Parameters
    ------------
    duration: :class:`int`
        The poll duration in minutes.
    id: Optional[`ID`]
        The poll's unique ID.
    voting_status: Optional[:class:`str`]
        The poll's voting status.
    end_date: Optional[:class:`str`]
        The poll's end date.


    .. versionadded:: 1.1.0
    """

    __slots__ = (
        "_id",
        "_voting_status",
        "_end_date",
        "_duration",
        "_options",
        "_raw_options",
    )

    def __init__(self, *, duration: int, **kwargs):
        self._id: Optional[ID] = kwargs.get("id", None)
        self._voting_status: Optional[str] = kwargs.get("voting_status", None)
        self._end_date = kwargs.get("end_date", None)
        self._duration = duration
        self._options = []
        self._raw_options = []
        super().__init__(self.id)

    def __repr__(self) -> str:
        return "Poll(id={0.id} voting_status={0.voting_status} duration={0.duration})".format(self)

    def __len__(self) -> int:
        return len(self.options)

    def add_option(self, *, label: str, **kwargs) -> Poll:
        """Add option to your Poll instance.

        .. note::
            For attaching a poll object you should use the label argument and only label as the api only need labels to make poll.

        Parameters
        ------------
        label: :class:`str`
            The option's label.
        position: :class:`int`
            The option's position.
        votes: :class:`int`
            The option's votes.

        Returns
        ---------
        :class:`Poll`
            This method return your :class:`Poll` instance.


        .. versionadded 1.3.5
        """
        data = {
            "position": kwargs.get("position") or 0,
            "label": label,
            "votes": kwargs.get("votes") or 0,
        }
        self._options.append(PollOption(**data))
        self._raw_options.append(data)
        return self

    @property
    def id(self) -> Optional[int]:
        """Optional[:class:`int`]: Return the poll's unique ID.

        .. versionadded:: 1.1.0
        """
        return int(self._id) if self._id else None

    @property
    def options(self) -> List[PollOption]:
        """List[:class:`PollOption`]: Return a list of :class:`PollOption`.

        .. versionadded:: 1.1.0
        """
        return self._options

    @property
    def raw_options(self) -> List[dict]:
        """List[:class:`dict`]: Return a list of raw poll option.

        .. versionadded:: 1.3.5
        """
        return self._raw_options

    @property
    def voting_status(self) -> str:
        """:class:`str`: Returns the current voting status of the poll.

        .. versionadded:: 1.1.0

        .. versionchanged:: 1.3.7
        """
        return self._voting_status

    @property
    def is_open(self) -> bool:
        """:class:`bool`: Return True if the poll is still open for voting, if it's closed it will returns False.

        .. versionadded:: 1.3.7
        """

        return self.voting_status == "open"

    @property
    def is_closed(self) -> bool:
        """:class:`bool`: Returns False if the poll is still open for voting, if it's closed it returns True.

        .. versionadded:: 1.3.7
        """

        return not self.is_open

    @property
    def duration(self) -> int:
        """:class:`int`: Return the poll duration in minutes.

        .. versionadded:: 1.3.5
        """
        return int(self._duration) if self._duration else None

    @property
    def end_date(self) -> Optional[datetime.datetime]:
        """Optional[:class:`datetime.datetime`]: Return the end date in datetime.datetime object.

        .. versionadded:: 1.1.0
        """
        return time_parse_todt(self._end_date) if self._end_date else None


class QuickReply:
    """Represents a quick_reply attachment in Direct Message.

    Parameters
    ------------
    type: :class:`str`
        The quick_reply's types, it must be and only 'options'


    .. versionadded:: 1.2.0
    """

    __slots__ = ("_type", "_options", "_raw_options")

    def __init__(self, type: str = "options"):
        self._type = type if type == "options" else "options"
        self._options: List[Option] = []
        self._raw_options: List[dict] = []

    def add_option(
        self,
        *,
        label: str,
        description: Optional[str] = None,
        metadata: Optional[str] = None,
    ) -> QuickReply:
        """Method for adding an option in your quick reply instance.

        Parameters
        ------------
        label: :class:`str`
            The option's label. Label text is returned as the user's message response, Must be less then 36 characters.
        description: :class:`str`
            The option's description. Description text displayed under label text. All options must have this property defined if property is present in any option. Text is auto-wrapped and will display on a max of two lines and supports n for controlling line breaks, Must be less then 72 characters.
        metadata: :class:`str`
            The option's metadata. Metadata that will be sent back in the webhook request, must be less then 1000 characters.

        Returns
        ---------
        :class:`QuickReply`
            This method return your :class:`QuickReply` instance.


        .. versionadded:: 1.2.0
        """

        self._raw_options.append({"label": label, "description": description, "metadata": metadata})
        self._options.append(Option(label=label, description=description, metadata=metadata))

        return self

    @property
    def type(self) -> str:
        """:class:`str`: Returns the QuickReply type.

        .. versionadded:: 1.5.0
        """
        return self._type

    @property
    def options(self) -> List[Option]:
        """List[:class:`Option`]: Returns a list of pre-made Option objects.

        .. versionadded:: 1.3.5
        """
        return self._options

    @property
    def raw_options(self) -> List[dict]:
        """List[:class:`dict`]: Returns the raw options.

        .. versionadded:: 1.2.0
        """
        return self._raw_options


class CTA:
    """Represents call-to-action attachment(CTA)
    You can use it in :meth:`User.send` via CTA kwarg. CTA will perform and action whenever a user "call" something, an example of this is buttons.

    .. versionadded:: 1.3.5
    """

    __slots__ = ("_buttons", "_raw_buttons")

    def __init__(self):
        self._buttons = []
        self._raw_buttons = []

    def add_button(
        self,
        *,
        label: str,
        url: str,
        type: Union[ButtonType, str] = ButtonType.web_url,
        tco_url: Optional[str] = None,
    ) -> CTA:
        """Add a button in your CTA instance.

        Parameters
        ------------
        label: :class:`str`
            The button's label, will be shown in the main text.
        url: :class:`str`
            A url that specified where to take you when you click the button, e.g you can take a user to someone's dm, a tweet, etc.
        type: :class:`ButtonType`
            The button's type, For now twitter only use web_url, if none specified the default type is ButtonType.web_url.
        tco_url: Optional[:class:`str`]
            The url in tco style.

        Returns
        ---------
        :class:`CTA`
            Returns your :class:`CTA` instance.


        .. versionadded:: 1.3.5
        """
        self._raw_buttons.append(
            {
                "type": type.value if isinstance(type, ButtonType) else type,
                "label": label,
                "url": url,
            }
        )

        self._buttons.append(Button(label, type, url, tco_url))
        return self

    @property
    def buttons(self) -> List[Button]:
        """List[:class:`Button`]: Returns a list of pre-made buttons object.

        .. versionadded:: 1.3.5
        """
        return self._buttons

    @property
    def raw_buttons(self) -> List[Dict]:
        """List[dict]: Returns the list of dictionaries filled with raw buttons.

        .. versionadded:: 1.3.5
        """
        return self._raw_buttons


class File:
    """Represents a File attachment for messages.

    Parameters
    ------------
    path: :class:`str`
        The file's path.
    dm_only: :class:`bool`
        Indicates if the file is use in dm only. Default to False.
    alt_text: Optional[:class:`str`]
        The image's alt text, if None specified the image won't have an alt text. Default to None.
    subtitle_language_code: :class:`str`
        The language code should be a BCP47 code (e.g. "en").
    subfile: :class:`SubFile`
        The subtitle's source file. Must be a .srt file with the correct formats.


    .. versionadded:: 1.3.5
    """

    __slots__ = (
        "__path",
        "__media_id",
        "__subfile",
        "__subfiles",
        "_total_bytes",
        "_mimetype",
        "dm_only",
        "alt_text",
    )

    def __init__(
        self,
        path: str,
        *,
        dm_only: bool = False,
        alt_text: Optional[str] = None,
        subfile: Optional[SubFile] = None,
        subfiles: Optional[List[SubFile]] = None,
    ):
        self.__path = path
        self.__media_id = None
        self.__subfile = subfile
        self.__subfiles = subfiles
        self._total_bytes = os.path.getsize(path) if isinstance(path, str) else os.path.getsize(path.name)
        self._mimetype = (
            guess_mimetype(open(path, "rb").read()) if isinstance(path, str) else guess_mimetype(path.read())
        )
        self.dm_only = dm_only
        self.alt_text = alt_text

    def __repr__(self) -> str:
        return "File(filename={0.filename})".format(self)

    @property
    def path(self) -> str:
        """:class:`str`: Returns the file's path.

        .. versionadded:: 1.3.5
        """
        return self.__path

    @property
    def media_id(self) -> Optional[int]:
        """Optional[:class:`int`]: Returns the file's media id. Returns None if the file was never uploaded.

        .. versionadded:: 1.5.0
        """
        return int(self.__media_id) if self.__media_id else self.__media_id

    @property
    def subfile_media_id(self) -> Optional[int]:
        """Optional[:class:`int`]: Returns the file's subtitle file's media id. Returns None if the subfile was never uploaded.

        .. versionadded:: 1.5.0
        """
        return int(self.subfile.media_id) if self.subfile.media_id else self.subfile.media_id

    @property
    def mimetype(self) -> str:
        """:class:`str`: Returns the file's mimetype.

        .. versionadded:: 1.3.5
        """
        return self._mimetype

    @property
    def filename(self) -> str:
        """:class:`str`: Returns the file's basename.

        .. versionadded:: 1.3.5
        """
        return self.path.name if isinstance(self.path, io.IOBase) else os.path.basename(self.path)

    @property
    def total_bytes(self) -> int:
        """:class:`int`: Returns an integer value that represents the size of the specified path in bytes.

        .. versionadded:: 1.3.5
        """
        return self._total_bytes

    @property
    def media_category(self) -> str:
        """:class:`str`: Returns the file's media category. e.g If its more tweet messages it can be TWEET_IMAGE if its in direct messages it will be dm_image.

        .. versionadded:: 1.3.5
        """
        startpoint = "TWEET_"
        mimetype = self._mimetype
        if mimetype in ("image/jpeg", "image/png"):
            return startpoint + "IMAGE" if not self.dm_only else "dm_image"
        elif mimetype == "image/gif":
            return startpoint + "GIF" if not self.dm_only else "dm_gif"
        elif mimetype == "video/mp4":
            return startpoint + "VIDEO" if not self.dm_only else "dm_video"

    @property
    def subfile(self) -> Optional[SubFile]:
        """Optional[:class:`SubFile`]: Returns the file's subfile.

        .. versionadded:: 1.5.0
        """
        return self.__subfile

    @property
    def subfiles(self) -> Optional[List[SubFile]]:
        """Optional[List[:class:`SubFile`]]: Returns a list of the file's subfiles.

        .. versionadded:: 1.5.0
        """
        return self.__subfiles


class SubFile(File):
    """Represents a subtitle File for :class:`File`. You can attach one subfile in :class:`File` via subfile arguments. This method inherits :class:`File`. Always remembers that twitter only supports srt extension file with the correct timestamps.


    .. versionadded:: 1.5.0
    """

    __slots__ = ("_language_code", "_language")

    def __init__(self, path: str, *, language_code: Optional[str]):
        self._language_code = language_code
        if self.language_code:
            fullname = LANGUAGES_CODES.get(language_code)
            if fullname:
                self._language = fullname
            else:
                raise pytwotException("Wrong language codes passed! Must be a BCP47 code (e.g. 'en')")
        super().__init__(path)

    def __repr__(self) -> str:
        return "SubFile(filename={0.filename} language={0.language} language_code={0.language_code})".format(self)

    @property
    def filename(self) -> str:
        """:class:`str`: Returns the SubFile's filename.

        .. versionadded:: 1.5.0
        """
        return self.path.name if isinstance(self.path, io.IOBase) else os.path.basename(self.path)

    @property
    def path(self) -> str:
        """:class:`str`: Returns the SubFile's path.

        .. versionadded:: 1.3.5
        """
        return self._File__path

    @property
    def language(self) -> str:
        """:class:`str`: Returns the SubFile's language.

        .. versionadded:: 1.5.0
        """
        return self._language

    @property
    def language_code(self) -> str:
        """:class:`str`: Returns the SubFile's language code.

        .. versionadded:: 1.5.0
        """
        return self._language_code

    @property
    def media_id(self) -> Optional[int]:
        """Optional[:class:`int`]: Returns the SubFile's media id. Returns None if the file was never uploaded.

        .. versionadded:: 1.5.0
        """
        return int(self._File__media_id) if self._File__media_id else self._File__media_id

    @property
    def total_bytes(self) -> int:
        """:class:`int`: Returns an integer value that represents the size of the specified path in bytes.

        .. versionadded:: 1.5.0
        """
        return self._total_bytes

    @property
    def mimetype(self) -> str:
        """:class:`str`: Returns the subfile's media category. This always returns 'text/srt'

        .. versionadded:: 1.5.0
        """
        return "text/srt"

    @property
    def media_category(self) -> str:
        """:class:`str`: Returns the subfile's media category. This always returns 'Subtitles'.

        .. versionadded:: 1.5.0
        """
        return "Subtitles"


class Geo(Comparable):
    """Represents the Geo location in twitter.
    You can use this as attachment in a tweet or for searching a location


    .. versionadded:: 1.3.5
    """

    __slots__ = ("_payload", "__bounding_box")

    def __init__(self, data: Dict[str, Any]):
        self._payload = data
        self.__bounding_box = self._payload.get("bounding_box")
        super().__init__(self.id)

    def __repr__(self) -> str:
        return "Geo(name={0.name} fullname={0.fullname} country={0.country} country_code={0.country_code} id={0.id})".format(
            self
        )

    @property
    def name(self) -> str:
        """:class:`str`: Returns the geo's name.

        .. versionadded:: 1.3.5
        """
        return self._payload.get("name")

    @property
    def id(self) -> str:
        """:class:`str`: Returns the geo's unique id.

        .. versionadded:: 1.3.5
        """
        return self._payload.get("id")

    @property
    def fullname(self) -> str:
        """:class:`str`: Returns the geo's fullname.

        .. versionadded:: 1.3.5
        """
        return self._payload.get("full_name")

    @property
    def type(self) -> str:
        """:class:`str`: Returns the geo's type.

        .. versionadded:: 1.3.5
        """
        return self._payload.get("place_type")

    @property
    def country(self) -> str:
        """:class:`str`: Returns the country where the geo is in.

        .. versionadded:: 1.3.5
        """
        return self._payload.get("country")

    @property
    def country_code(self) -> str:
        """:class:`str`: Returns the country's code where the geo is in.

        .. versionadded:: 1.3.5
        """
        return self._payload.get("country_code")

    @property
    def centroid(self) -> str:
        """:class:`str`: Returns the geo's centroid.

        .. versionadded:: 1.3.5
        """
        return self._payload.get("centroid")

    @property
    def bounding_box_type(self) -> str:
        """:class:`str`: Returns the geo's bounding box type.

        .. versionadded:: 1.3.5
        """
        if self.__bounding_box:
            return self.__bounding_box.get("type")
        return None

    @property
    def coordinates(self) -> List[str]:
        """List[:class:`str`]: Returns a list of coordinates where the geo's located.

        .. versionadded:: 1.3.5
        """
        if self.__bounding_box:
            return self.__bounding_box.get("coordinates")
        return None


class CustomProfile(Comparable):
    """Represents a CustomProfile attachments that allow a Direct Message author to present a different identity than that of the Twitter account being used.

    .. versionadded:: 1.3.5
    """

    __slots__ = ("_name", "_id", "_timestamp", "_media")

    def __init__(
        self,
        name: str,
        id: ID,
        timestamp: ID,
        media: Dict[str, Any],
    ):
        self._name = name
        self._id = id
        self._timestamp = timestamp
        self._media = Media(media)
        super().__init__(self.id)

    def __repr__(self) -> str:
        return "CustomProfile(name={0.name} id={0.id} media_id={0.media_id})".format(self)

    @property
    def name(self) -> str:
        """:class:`str`: The author's custom name.

        .. versionadded:: 1.3.5
        """
        return self._name

    @property
    def id(self) -> int:
        """:class:`int`: The custom profile unique ID.

        .. versionadded:: 1.3.5
        """
        return int(self._id)

    @property
    def created_at(self) -> datetime.datetime:
        """:class:`datetime.datetime`: Returns a datetime.datetime object with the CustomProfile created time.

        .. versionadded:: 1.3.5
        """
        return datetime.datetime.fromtimestamp(int(self._timestamp) / 1000)

    @property
    def media(self) -> Media:
        """:class:`Media`: Returns the media object.

        .. versionadded:: 1.3.5
        """
        return self._media



"""
The MIT License (MIT)

Copyright (c) 2021-present UnrealFar & TheGenocides
Copyright (c) 2023-present Sengolda

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import io
import datetime
from typing import TYPE_CHECKING, Any, Dict, Optional, Union, List

from .constants import (
    TWEET_EXPANSION,
    LIST_EXPANSION,
    PINNED_TWEET_EXPANSION,
    MEDIA_FIELD,
    PLACE_FIELD,
    POLL_FIELD,
    TWEET_FIELD,
    USER_FIELD,
    LIST_FIELD,
)
from .relations import RelationFollow
from .utils import time_parse_todt, convert
from .dataclass import UserSettings, Location
from .paginations import UserPagination, TweetPagination, ListPagination
from .list import List as TwitterList
from .objects import Comparable

if TYPE_CHECKING:
    from .message import DirectMessage
    from .http import HTTPClient
    from .type import ID
    from .tweet import Tweet
    from .enums import Timezone
    from .attachments import Geo, CTA, CustomProfile, File, QuickReply


class User(Comparable):
    """Represents a user in Twitter.
    User is an identity in twitter, its very interactive. Can send message, post a tweet, and even send messages to other user through Dms.


    .. describe:: x == y

        Check if one user id is equal to another.


    .. describe:: x != y

        Check if one user id is not equal to another.


    .. describe:: str(x)

        Get the user's name.


    .. versionadded: 1.0.0
    """

    __slots__ = (
        "__original_payload",
        "_includes",
        "_payload",
        "http_client",
        "_public_metrics",
    )

    def __init__(self, data: Dict[str, Any], http_client: Optional[HTTPClient] = None) -> None:
        self.__original_payload = data
        self._includes = self.__original_payload.get("includes")
        self._payload = self.__original_payload.get("data") or self.__original_payload
        self.http_client = http_client
        self._public_metrics = self._payload.get("public_metrics", None) or self.__original_payload.get(
            "public_metrics"
        )
        super().__init__(self.id)

    def __str__(self) -> str:
        return self.username

    def __repr__(self) -> str:
        return "User(name={0.name} username={0.username} id={0.id})".format(self)

    @property
    def name(self) -> str:
        """:class:`str`: Return the user's name.

        .. versionadded: 1.0.0
        """
        return self._payload.get("name")

    @property
    def username(self) -> str:
        """:class:`str`: Return the user's username.

        .. versionadded: 1.0.0
        """
        return self._payload.get("username")

    @property
    def id(self) -> int:
        """:class:`int`: Return the user's id.

        .. versionadded: 1.0.0
        """
        return int(self._payload.get("id"))

    @property
    def description(self) -> str:
        """:class:`str`: Return the user's description.

        .. versionadded: 1.0.0
        """
        return self._payload.get("description")

    @property
    def bio(self) -> str:
        """:class:`str`: an alias to :meth:`User.description`.

        .. versionadded: 1.0.0
        """
        return self.description

    @property
    def mention(self) -> str:
        """:class:`str`: Return the user mention format.

        .. versionadded: 1.5.0
        """
        return "@" + self.username

    @property
    def pinned_tweet(self) -> Optional[Tweet]:
        """Optional[:class:`Tweet`]: Returns the user's pinned tweet. Returns None if the user dont have one.

        .. versionadded:: 1.5.0
        """
        from .tweet import Tweet  # Avoid circular import error.

        if self._includes and self._includes.get("tweets"):
            data = {}
            data["data"] = self._includes.get("tweets")[0]
            data["includes"] = {}
            data["includes"]["users"] = [self.__original_payload]
            return Tweet(data, http_client=self.http_client)
        return None

    @property
    def url(self) -> Optional[str]:
        """:class:`str`: Return url that associated with the user profile.

        .. versionadded: 1.0.0
        """
        return self._payload.get("url", None)

    @property
    def profile_url(self) -> str:
        """:class:`str`: Return the user's profile url.

        .. versionadded: 1.0.0
        """
        return f"https://twitter.com/{self.username}"

    @property
    def profile_image_url(self) -> Optional[str]:
        """Optional[:class:`str`] Return the user profile image url.

        .. versionadded: 1.0.0
        """
        return self._payload.get("profile_image_url", None)

    @property
    def verified(self) -> bool:
        """:class:`bool`: Return True if the user is verified account, else False.

        .. versionadded: 1.0.0
        """
        return self._payload.get("verified")

    @property
    def protected(self) -> bool:
        """:class:`bool`: Return True if the user is protected, else False.

        .. versionadded: 1.0.0
        """
        return self._payload.get("protected")

    @property
    def private(self) -> bool:
        """:class:`bool`: An alias to :meth:`User.protected`.

        .. versionadded: 1.3.5
        """
        return self.protected

    @property
    def location(self) -> Optional[str]:
        """:class:`str`: Return the user's location.

        .. versionadded: 1.0.0
        """
        return self._payload.get("location", None)

    @property
    def created_at(self) -> datetime.datetime:
        """Optional[:class:`datetime.datetime`]: Returns a datetime.datetime object with the user's account date.

        .. versionadded: 1.0.0
        """
        try:
            timestamp = int(self._payload.get("created_at"))
            return datetime.datetime.fromtimestamp(timestamp / 1000)
        except ValueError:
            return time_parse_todt(self._payload.get("created_at"))

    @property
    def follower_count(self) -> Optional[int]:
        """Optional[:class:`int`]: Returns total of followers that the user has.

        .. versionadded: 1.1.0
        """
        return convert(self._public_metrics.get("follower_count"), int)

    @property
    def following_count(self) -> Optional[int]:
        """Optional[:class:`int`]: Returns total of following that the user has.

        .. versionadded: 1.1.0
        """
        return convert(self._public_metrics.get("following_count"), int)

    @property
    def tweet_count(self) -> Optional[int]:
        """Optional[:class:`int`]: Returns total of tweet that the user has.

        .. versionadded: 1.1.0
        """
        return convert(self._public_metrics.get("tweet_count"), int)

    @property
    def listed_count(self) -> Optional[int]:
        """Optional[:class:`int`]: Returns total of listed that the user has.

        .. versionadded: 1.1.0
        """
        return convert(self._public_metrics.get("listed_count"), int)

    def send(
        self,
        text: str,
        *,
        file: Optional[File] = None,
        custom_profile: Optional[CustomProfile] = None,
        quick_reply: Optional[QuickReply] = None,
        cta: Optional[CTA] = None,
    ) -> DirectMessage:
        """Send a message to the user.

        Parameters
        ------------
        text: :class:`str`
            The text that will be send to that user.
        file: Optional[:class:`File`]
            Represents a single file attachment. It could be an image, gif, or video. It also have to be an instance of pytwot.File.
        custom_profile: Optional[:class:`CustomProfile`]
            The custom profile attachment.
        quick_reply: Optional[:class:`QuickReply`]
            The QuickReply attachment that will be send to a user.
        cta: Optional[:class:`CTA`]
            cta or call-to-actions is use to make an action whenever a user 'call' something, a quick example is buttons.

        Returns
        ---------
        :class:`DirectMessage`
            This method return a :class:`DirectMessage` object.


        .. versionadded:: 1.1.0
        """
        return self.http_client.send_message(
            self.id,
            text,
            file=file,
            custom_profile=custom_profile,
            quick_reply=quick_reply,
            cta=cta,
        )

    def follow(self) -> RelationFollow:
        """Follows the User.

        Returns
        ---------
        :class:`RelationFollow`
            This method return a :class:`RelationFollow` object.


        .. versionadded:: 1.1.0
        """
        my_id = self.http_client.access_token.partition("-")[0]
        res = self.http_client.request(
            "POST",
            "2",
            f"/users/{my_id}/following",
            json={"target_user_id": str(self.id)},
            auth=True,
        )
        return RelationFollow(res)

    def unfollow(self) -> RelationFollow:
        """Unfollows the User.

        Returns
        ---------
        :class:`RelationFollow`
            This method return a :class:`RelationFollow` object.


        .. versionadded:: 1.1.0
        """
        my_id = self.http_client.access_token.partition("-")[0]
        res = self.http_client.request("DELETE", "2", f"/users/{my_id}/following/{self.id}", auth=True)
        return RelationFollow(res)

    def block(self) -> None:
        """Blocks the user.

        .. versionadded:: 1.2.0
        """
        my_id = self.http_client.access_token.partition("-")[0]
        self.http_client.request(
            "POST",
            "2",
            f"/users/{my_id}/blocking",
            json={"target_user_id": str(self.id)},
            auth=True,
        )

    def unblock(self) -> None:
        """Unblocks the user.

        .. versionadded:: 1.2.0
        """
        my_id = self.http_client.access_token.partition("-")[0]
        self.http_client.request("DELETE", "2", f"/users/{my_id}/blocking/{self.id}", auth=True)

    def mute(self) -> None:
        """Mutes the user.

        .. versionadded:: 1.2.5
        """
        my_id = self.http_client.access_token.partition("-")[0]
        self.http_client.request(
            "POST",
            "2",
            f"/users/{my_id}/muting",
            json={"target_user_id": str(self.id)},
            auth=True,
        )

    def unmute(self) -> None:
        """Unmutes the user.

        .. versionadded:: 1.2.5
        """
        my_id = self.http_client.access_token.partition("-")[0]
        self.http_client.request("DELETE", "2", f"/users/{my_id}/muting/{self.id}", auth=True)

    def report(self, *, block: bool = True):
        """Reports the user as a spam account to twitter.

        Parameters
        -----------
        block: :class:`bool`
            Indicates to perform a block action to the user if set to True. Default to True.


        .. versionadded:: 1.3.5
        """
        self.http_client.request(
            "POST",
            "1.1",
            "/users/report_spam.json",
            params={"user_id": str(self.id), "perform_block": block},
        )

    def trigger_typing(self):
        """Trigger the typing animation in the user's dm. W ill stop when the client decide to send a message when the typing animation is playing or in couple of seconds.

        .. versionadded:: 1.3.5
        """
        self.http_client.request(
            "POST",
            "1.1",
            "/direct_messages/indicate_typing.json",
            params={"recipient_id": str(self.id)},
            auth=True,
        )

    def fetch_followers(self) -> Optional[UserPagination]:
        """Fetches users from the user's followers list then paginate it .

        Returns
        ---------
        Optional[:class:`UserPagination`]
            This method returns a :class:`UserPagination` object.


        .. versionadded:: 1.3.5
        """
        following = self.http_client.request(
            "GET",
            "2",
            f"/users/{self.id}/followers",
            params={
                "expansions": PINNED_TWEET_EXPANSION,
                "user.fields": USER_FIELD,
                "tweet.fields": TWEET_FIELD,
            },
        )

        if not following:
            return []

        return UserPagination(
            following,
            endpoint_request=f"/users/{self.id}/followers",
            http_client=self.http_client,
            params={
                "expansions": PINNED_TWEET_EXPANSION,
                "user.fields": USER_FIELD,
                "tweet.fields": TWEET_FIELD,
            },
        )

    def fetch_following(self) -> Optional[UserPagination]:
        """Fetches users from the user's following list then paginate it.

        Returns
        ---------
        Optional[:class:`UserPagination`]
            This method returns a :class:`UserPagination` object.


        .. versionadded:: 1.3.5
        """
        following = self.http_client.request(
            "GET",
            "2",
            f"/users/{self.id}/following",
            params={
                "expansions": PINNED_TWEET_EXPANSION,
                "user.fields": USER_FIELD,
                "tweet.fields": TWEET_FIELD,
            },
        )

        if not following:
            return []

        return UserPagination(
            following,
            endpoint_request=f"/users/{self.id}/following",
            http_client=self.http_client,
            params={
                "expansions": PINNED_TWEET_EXPANSION,
                "user.fields": USER_FIELD,
                "tweet.fields": TWEET_FIELD,
            },
        )

    def fetch_timelines(
        self,
        *,
        start_time: Optional[datetime.datetime] = None,
        end_time: Optional[datetime.datetime] = None,
        since_id: Optional[ID] = None,
        until_id: Optional[ID] = None,
        mentioned: bool = False,
        exclude: Optional[str] = None,
    ) -> Union[TweetPagination, list]:
        """Fetches the user timelines, this can be timelines where the user got mention or a normal tweet timelines.

        Parameters
        ------------
        start_time: Optional[:class:`datetime.datetime`]
            This will make sure the tweets created datetime is after that specific time.
        end_time: Optional[:class:`datetime.datetime`]
            This will make sure the tweets created datetime is before that specific time.
        since_id: Optional[`ID`]
            Returns results with a Tweet ID greater than (that is, more recent than) the specified 'since' Tweet ID. Only the 3200 most recent Tweets are available. The result will exclude the since_id. If the limit of Tweets has occurred since the since_id, the since_id will be forced to the oldest ID available.
        until_id: Optional[`ID`]
            Returns results with a Tweet ID less less than (that is, older than) the specified 'until' Tweet ID. Only the 3200 most recent Tweets are available. The result will exclude the until_id. If the limit of Tweets has occurred since the until_id, the until_id will be forced to the most recent ID available.
        mentioned: :class:`bool`
            Indicates if only mentioned timelines return if set to True, else it will returns a normal tweet timelines. Default to False.
        exclude: :class:`str`
            Specified which tweet type should not be returns, you can set it to:'retweets,replies' or 'retweets' or 'replies'.

        Returns
        ---------
        Union[:class:`TweetPagination`, :class:`list`]
            This method returns a :class:`TweetPagination` objects or an empty :class:`list` if none founded.


        .. versionadded:: 1.3.5
        """
        if (
            not isinstance(start_time, datetime.datetime)
            and start_time
            or not isinstance(end_time, datetime.datetime)
            and end_time
        ):
            raise ValueError("start_time or end_time must be a datetime object!")

        params = {
            "expansions": TWEET_EXPANSION,
            "user.fields": USER_FIELD,
            "media.fields": MEDIA_FIELD,
            "place.fields": PLACE_FIELD,
            "poll.fields": POLL_FIELD,
            "tweet.fields": TWEET_FIELD,
        }

        if start_time:
            params["start_time"] = start_time.isoformat()
        if end_time:
            params["end_time"] = end_time.isoformat()
        if since_id:
            params["since_id"] = str(since_id)
        if until_id:
            params["until_id"] = str(until_id)
        if exclude:
            params["exclude"] = exclude

        res = self.http_client.request(
            "GET",
            "2",
            f"/users/{self.id}/tweets" if not mentioned else f"/users/{self.id}/mentions",
            params=params,
        )

        if not res:
            return []
        return TweetPagination(
            res,
            endpoint_request=f"/users/{self.id}/tweets" if not mentioned else f"/users/{self.id}",
            http_client=self.http_client,
            params=params,
        )

    def fetch_liked_tweets(self) -> TweetPagination:
        """Fetches tweets that's been liked by the user.

        .. versionadded:: 1.5.0
        """
        params = {
            "expansions": TWEET_EXPANSION,
            "user.fields": USER_FIELD,
            "media.fields": MEDIA_FIELD,
            "place.fields": PLACE_FIELD,
            "poll.fields": POLL_FIELD,
            "tweet.fields": TWEET_FIELD,
        }

        res = self.http_client.request(
            "GET",
            "2",
            f"/users/{self.id}/liked_tweets",
            params=params,
        )

        if not res:
            return []
        return TweetPagination(
            res,
            endpoint_request=f"/users/{self.id}/liked_tweets",
            http_client=self.http_client,
            params=params,
        )

    def fetch_pinned_tweet(self) -> Optional[Tweet]:
        """Fetches the user's pinned tweet, consider using this method if :meth:`User.pinned_tweet` returns None.

        Returns
        ---------
        Optional[:class:`Tweet`]
            This method returns a :class:`Tweet` object.


        .. versionadded: 1.1.3
        """
        id = self._payload.get(PINNED_TWEET_EXPANSION)
        return self.http_client.fetch_tweet(int(id)) if id else None

    def fetch_lists(self) -> Optional[List[TwitterList]]:
        """Fetches the user's lists

        Returns
        ---------
        Optional[List[:class:`List`]]
            This method returns a list of :class:`List` objects.


        .. versionadded:: 1.5.0
        """
        params = {
            "expansions": LIST_EXPANSION,
            "list.fields": LIST_FIELD,
            "user.fields": USER_FIELD,
        }

        res = self.http_client.request("GET", "2", f"/users/{self.id}/owned_lists", params=params)

        if not res:
            return []

        return ListPagination(
            res,
            endpoint_request=f"/users/{self.id}/owned_lists",
            http_client=self.http_client,
            params=params,
        )

    def fetch_pinned_lists(self) -> Optional[List[TwitterList]]:
        """Fetches the user's pinned lists, returns an empty list if not found

        Returns
        ---------
        Optional[List[:class:`List`]]:
            This method returns a list of :class:`List` objects.


        .. versionadded:: 1.5.0
        """
        res = self.http_client.request(
            "GET",
            "2",
            f"/users/{self.id}/pinned_lists",
            auth=True,
            params={
                "expansions": LIST_EXPANSION,
                "user.fields": USER_FIELD,
                "list.fields": LIST_FIELD,
            },
        )
        if not res:
            return None

        for data in res["data"]:
            self.http_client.payload_parser.insert_object_author(data, self)

        return [TwitterList(data, http_client=self.http_client) for data in res["data"]]

    def fetch_list_memberships(self) -> Union[ListPagination, List]:
        """Fetches all lists the user is a member of.

        Returns
        ---------
        Optional[List[:class:`List`]]
            This method returns a list of :class:`List` objects.


        .. versionadded:: 1.5.0
        """
        params = {
            "expansions": LIST_EXPANSION,
            "list.fields": LIST_FIELD,
            "user.fields": USER_FIELD,
        }

        res = self.http_client.request("GET", "2", f"/users/{self.id}/list_memberships", params=params)

        if not res:
            return []

        return ListPagination(
            res,
            endpoint_request=f"/users/{self.id}/list_memberships",
            http_client=self.http_client,
            params=params,
        )

    def fetch_followed_lists(self) -> Union[ListPagination, List]:
        """Fetches the user's followed lists.


        .. versionadded:: 1.5.0
        """
        params = {
            "expansions": LIST_EXPANSION,
            "list.fields": LIST_FIELD,
            "user.fields": USER_FIELD,
        }

        res = self.http_client.request("GET", "2", f"/users/{self.id}/followed_lists", params=params)

        if not res:
            return []

        return ListPagination(
            res,
            endpoint_request=f"/users/{self.id}/followed_lists",
            http_client=self.http_client,
            params=params,
        )


class ClientAccount(User):
    """Represents the client's account. This inherits :class:`User` object. This class unlocks methods that you can only use for the authenticated user (the client).

    .. versionadded:: 1.5.0
    """

    def update_settings(
        self,
        *,
        lang: Optional[str] = None,
        enabled_sleep_time: Optional[bool] = None,
        start_sleep_time: Optional[datetime.datetime] = None,
        end_sleep_time: Optional[datetime.datetime] = None,
        timezone: Optional[Timezone] = None,
        location: Optional[Location, int] = None,
    ):
        """Updates the user settings.

        Parameters
        ------------
        lang: Optional[:class:`str`]
            The new language replacing the old one.
        enabled_sleep_time: Optional[:class:`bool`]
            Indicates to enabled sleep time.
        start_sleep_time: Optional[:class:`int`]
            The hour that sleep time should begin if it is enabled. Must be an instance of datetime.datetime.
        end_sleep_time: Optional[:class:`int`]
            The hour that sleep time should end if it is enabled. Must be an instance of datetime.datetime.
        timezone: Optional[:class:`Timezone`]
            The new timezone replacing the old one. Must be an instance of :class:`Timezone` (e.g :attr:`Timezone.jakarta` or :attr:`Timezone.paris`)
        location: Optional[:class:`Location`, :class:`int`]
            The Yahoo! Where On Earth ID to use as the user's default trend location. Global information is available by using 1 as the WOEID. Must be an instance of :class:`Location` or the woeid in :class:`int`.


        .. versionadded:: 1.5.0
        """
        res = self.http_client.request(
            "POST",
            "1.1",
            "/account/settings.json",
            params={
                "sleep_time_enabled": enabled_sleep_time,
                "start_sleep_time": start_sleep_time,
                "end_sleep_time": end_sleep_time,
                "time_zone": timezone.value if isinstance(timezone, Timezone) else timezone,
                "trend_location_woeid": location.woeid if isinstance(location, Location) else location,
                "lang": lang,
            },
            auth=True,
        )
        if res.get("sleep_time"):
            self.http_client.payload_parser.parse_sleep_time_payload(res)

        if res.get("location"):
            self.http_client.payload_parser.parse_trend_location_payload(res)

        if res.get("time_zone"):
            self.http_client.payload_parser.parse_time_zone_payload(res)

        return UserSettings(**res)

    def fetch_settings(self):
        """Fetches the user settings.


        .. versionadded:: 1.5.0
        """
        res = self.http_client.request("GET", "1.1", "/account/settings.json", auth=True)
        if res.get("sleep_time"):
            self.http_client.payload_parser.parse_sleep_time_payload(res)

        if res.get("location"):
            self.http_client.payload_parser.parse_trend_location_payload(res)

        if res.get("time_zone"):
            self.http_client.payload_parser.parse_time_zone_payload(res)

        return UserSettings(**res)

    def update_profile(
        self,
        *,
        name: Optional[str] = None,
        description: Optional[str] = None,
        image: Optional[File] = None,
        location: Optional[Union[Geo, str]] = None,
        profile_url: Optional[str] = None,
        profile_link_color: Optional[ID] = None,
    ) -> ClientAccount:
        """Updates the client profile information from the given arguments.

        Parameters
        ------------
        name: Optional[:class:`str`]
            The new name replacing the client's oldname. Note that this isn't going to update the username.
        description: Optional[:class:`str`]
            The new description that you want to replace with the old version.
        image: Optional[:class:`File`]
            The new profile image that you want to replace with the old version. Must be an instance of :class:`File`.
        location: Optional[:class:`Geo`]
            The new location you want to replace with the old version. Must be an instance of :class:`Geo` or the fullname of geo. You use :meth:`Geo.fullname` to get the fullname.
        profile_url: Optional[:class:`str`]
            URL associated with the profile. Will be prepended with http:// if not present.
        profile_link_color: Optional[:class:`str`]
            Sets a hex value that controls the color scheme of links used on the authenticating user's profile page on twitter.com. This must be a valid hexadecimal value, and may be either three or six characters (ex: F00 or FF0000 in string). If you specified integer instead of string, the process will use hex() function to change the int value to a hex value.

        Returns
        ---------
        :class:`ClientAccount`
            Returns a updated client's account object.


        .. versionadded:: 1.5.0
        """
        if image:
            path = image.path
            if isinstance(path, io.IOBase):
                self.http_client.request(
                    "POST",
                    "1.1",
                    "/account/update_profile_image.json",
                    files={"image": path.read(4 * 1024 * 1024)},
                    auth=True,
                )
            else:
                self.http_client.request(
                    "POST",
                    "1.1",
                    "/account/update_profile_image.json",
                    files={"image": open(path, "rb").read(4 * 1024 * 1024)},
                    auth=True,
                )

        if isinstance(profile_link_color, int):
            profile_link_color = hex(profile_link_color).replace("0x", "", 1)

        res = self.http_client.request(
            "POST",
            "1.1",
            "/account/update_profile.json",
            params={
                "name": name,
                "description": description,
                "location": location.fullname if isinstance(location, Geo) else location,
                "url": profile_url,
                "profile_link_color": str(profile_link_color),
            },
            auth=True,
        )
        data = self.http_client.payload_parser.parse_user_payload(res)
        updated_account = ClientAccount(data, http_client=self.http_client)
        self = updated_account
        return self

    def update_profile_banner(
        self,
        *,
        banner: File,
        width: int = 0,
        height: int = 0,
        offset_left: int = 0,
        offset_top: int = 0,
    ) -> None:
        """Updates the profile banner.

        Parameters
        ------------
        banner: :class:`File`
            The new banner to replace the old one. Must be an instance of :class:`File`.
        width: :class:`int`
            The width of the preferred section of the image being uploaded in pixels. Use with height , offset_left , and offset_top to select the desired region of the image to use.
        height: :class:`int`
            The height of the preferred section of the image being uploaded in pixels. Use with width , offset_left , and offset_top to select the desired region of the image to use.
        offset_left: :class:`int`
            The number of pixels by which to offset the uploaded image from the left. Use with height , width , and offset_top to select the desired region of the image to use.
        offset_top: :class:`int`
            The number of pixels by which to offset the uploaded image from the top. Use with height , width , and offset_left to select the desired region of the image to use.


        .. versionadded:: 1.5.0
        """
        path = banner.path

        if isinstance(path, io.IOBase):
            self.http_client.request(
                "POST",
                "1.1",
                "/account/update_profile_banner.json",
                params={
                    "width": width,
                    "height": height,
                    "offset_left": offset_left,
                    "offset_top": offset_top,
                },
                files={"banner": path.read(4 * 1024 * 1024)},
                auth=True,
            )
        else:
            self.http_client.request(
                "POST",
                "1.1",
                "/account/update_profile_banner.json",
                params={
                    "width": width,
                    "height": height,
                    "offset_left": offset_left,
                    "offset_top": offset_top,
                },
                files={"banner": open(path, "rb").read(4 * 1024 * 1024)},
                auth=True,
            )
        return None

    def remove_profile_banner(self):
        """Remove the user profile banner.


        .. versionadded:: 1.5.0
        """
        self.http_client.request("POST", "1.1", "/account/remove_profile_banner.json", auth=True)

    def fetch_blockers(self) -> Optional[UserPagination]:
        """Fetches users from the client's block list then paginate it.

        Returns
        ---------
        Optional[:class:`UserPagination`]
            This method returns a :class:`UserPagination` object.


        .. versionadded:: 1.5.0
        """
        blockers = self.http_client.request(
            "GET",
            "2",
            f"/users/{self.id}/blocking",
            params={
                "expansions": PINNED_TWEET_EXPANSION,
                "user.fields": USER_FIELD,
                "tweet.fields": TWEET_FIELD,
            },
            auth=True,
        )

        if not blockers:
            return []

        return UserPagination(
            blockers,
            endpoint_request=f"/users/{self.id}/blocking",
            http_client=self.http_client,
            params={
                "expansions": PINNED_TWEET_EXPANSION,
                "user.fields": USER_FIELD,
                "tweet.fields": TWEET_FIELD,
            },
        )

    def fetch_muters(self) -> Optional[UserPagination]:
        """Fetches users from the client's mute list then paginate it.

        Returns
        ---------
        Optional[:class:`UserPagination`]
            This method returns a :class:`UserPagination` object.


        .. versionadded:: 1.5.0
        """
        muters = self.http_client.request(
            "GET",
            "2",
            f"/users/{self.id}/muting",
            params={
                "expansions": PINNED_TWEET_EXPANSION,
                "user.fields": USER_FIELD,
                "tweet.fields": TWEET_FIELD,
            },
            auth=True,
        )

        if not muters:
            return []

        return UserPagination(
            muters,
            endpoint_request=f"/users/{self.id}/muting",
            http_client=self.http_client,
            params={
                "expansions": PINNED_TWEET_EXPANSION,
                "user.fields": USER_FIELD,
                "tweet.fields": TWEET_FIELD,
            },
        )

    def fetch_bookmarks(self) -> Optional[TweetPagination]:
        """"""
        params = {
            "expansions": TWEET_EXPANSION,
            "user.fields": USER_FIELD,
            "tweet.fields": TWEET_FIELD,
            "media.fields": MEDIA_FIELD,
            "place.fields": PLACE_FIELD,
            "poll.fields": POLL_FIELD,
        }

        tweets = self.http_client.request(
            "GET",
            "2",
            f"/users/{self.id}/bookmarks",
            params=params,
            auth=True,
        )

        if not tweets:
            return []

        return UserPagination(
            tweets,
            endpoint_request=f"/users/{self.id}/bookmarks",
            http_client=self.http_client,
            params=params,
        )



"""
The MIT License (MIT)

Copyright (c) 2021-present UnrealFar & TheGenocides
Copyright (c) 2023-present Sengolda

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


from __future__ import annotations

import datetime
from typing import Any, Optional, TYPE_CHECKING
from dateutil import parser

if TYPE_CHECKING:
    from .type import ID


def convert(o: object, annotations: Any):
    try:
        return annotations(o)
    except (ValueError, TypeError):
        return object


def guess_mimetype(byts: bytes):
    if byts[6:10] == b"\x1a\n\x00\x00":
        return "image/png"

    elif byts[6:10] == b"JFIF":
        return "image/jpeg"

    elif byts[6:10] == b"ypis":
        return "video/mp4"

    elif byts.startswith((b"\x47\x49\x46\x38\x37\x61", b"\x47\x49\x46\x38\x39\x61")):
        return "image/gif"

    else:
        return "text/plain"


def time_parse_todt(date: Optional[Any]) -> datetime.datetime:
    """Parse time return from twitter to datetime object!

    Returns
    ---------
    :class:`datetime.datetime`


    .. versionadded: 1.1.3
    """
    date = str(parser.parse(date))
    y, mo, d = date.split("-")
    h, mi, s = date.split(" ")[1].split("+")[0].split(":")

    return datetime.datetime(
        year=int(y),
        month=int(mo),
        day=int(d.split(" ")[0]),
        hour=int(h),
        minute=int(mi),
        second=int(s),
    )


def compose_tweet(text: Optional[str] = None) -> str:
    """Make a link that let's you compose a tweet

    Parameters
    ------------
    text: :class:`str`
        The pre-populated text in the tweet. If none specified the user has to write their own message.


    Returns
    ---------
    :class:`str`


    .. versionadded: 1.3.5
    """
    if text:
        text = text.replace(" ", "%20")
    return (
        "https://twitter.com/intent/tweet"
        if not text
        else f"https://twitter.com/intent/tweet" + f"?text={text}"
        if text
        else f"https://twitter.com/intent/tweet"
    )


def compose_user_action(user_id: str, action: str, text: str = None):
    """Make a link that let's you interact with a user with certain actions.

    Parameters
    ------------
    user_id: :class:`str`
        The user's id.
    action: :class:`str`
        The action you are going to perform to the user. Must be either "follow" or "dm"
    text: :class:`str`
        The pre-populated text for the dm action.


    Returns
    ---------
    :class:`str`


    .. versionadded: 1.3.5
    """
    if action.lower() not in ("follow", "dm"):
        return TypeError("Action must be either 'follow' or 'dm'")
    if text:
        text = text.replace(" ", "%20")
    return (
        f"https://twitter.com/intent/user?user_id={user_id}"
        if action.lower() == "follow"
        else f"https://twitter.com/messages/compose?recipient_id={user_id}" + f"?text={text}"
        if text
        else f"https://twitter.com/messages/compose?recipient_id={user_id}"
    )


def compose_tweet_action(tweet_id: ID, action: str = None):
    """Make a link that let's you interact with a tweet with certain actions.

    Parameters
    ------------
    tweet_id: `ID`
        The tweet id you want to compose.
    action: :class:`str`
        The action that's going to get perform when you click the link. Must be either "retweet", "like", or "reply"

    Returns
    ---------
    :class:`str`


    .. versionadded: 1.3.5
    """
    if action.lower() not in ("retweet", "like", "reply"):
        return TypeError("Action must be either 'retweet', 'like', or 'reply'")
    return (
        f"https://twitter.com/intent/{action}?tweet_id={tweet_id}"
        if action != "reply"
        else f"https://twitter.com/intent/tweet?in_reply_to={tweet_id}"
    )



"""
The MIT License (MIT)

Copyright (c) 2021-present UnrealFar & TheGenocides
Copyright (c) 2023-present Sengolda

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, Optional

from .type import ID, Payload
from .utils import time_parse_todt
from .paginations import UserPagination, TweetPagination
from .constants import TWEET_EXPANSION, USER_FIELD, TWEET_FIELD, PINNED_TWEET_EXPANSION
from .relations import RelationUpdate, RelationDelete, RelationPin
from .objects import Comparable
from .relations import RelationFollow

if TYPE_CHECKING:
    from .http import HTTPClient
    from .user import User

__all__ = ("List", "_CopyList")


class List(Comparable):
    """Represents a Twitter List object

    A Twitter List is a curated group of accounts. Create one or subscribe to a list created by others to streamline your timeline.

    .. versionadded:: 1.5.0
    """

    __slots__ = ("__original_payload", "_payload", "http_client")

    def __init__(self, data: Payload, *, http_client: HTTPClient):
        self.__original_payload = data
        self._payload = self.__original_payload.get("data") or self.__original_payload
        self.http_client = http_client
        super().__init__(self.id)

    def __repr__(self) -> str:
        return "List(name={0.name} id={0.id} description={0.description} owner={0.owner!r})".format(self)

    @property
    def name(self) -> str:
        """`str`: Returns the list's name

        .. versionadded:: 1.5.0
        """
        return self._payload.get("name")

    @property
    def id(self) -> ID:
        """`ID`: Returns the list's id

        .. versionadded:: 1.5.0
        """
        return int(self._payload.get("id"))

    @property
    def description(self) -> Optional[str]:
        """Optional[:class:`str`]: Returns the list's description or None if the list doesn't have one.

        .. versionadded:: 1.5.0
        """
        description = self._payload.get("description")
        return description if description else None

    @property
    def owner(self) -> Optional[User]:
        """:class:`User`: Returns the list's owner in a user object

        .. versionadded:: 1.5.0
        """
        from .user import User  # Avoid circular import error

        user_data = self.__original_payload.get("includes").get("users")
        if user_data:
            return User(user_data[0], http_client=self.http_client)
        return None

    @property
    def created_at(self) -> datetime.datetime:
        """:class:`datetime.datetime`: Returns a datetime.datetime object with the list created creation time.

        .. versionadded:: 1.5.0
        """
        return time_parse_todt(self._payload.get("created_at"))

    @property
    def private(self) -> bool:
        """:class:`bool`: Returns True if the list is private else False.

        .. versionadded:: 1.5.0
        """
        return self._payload.get("private")

    @property
    def member_count(self) -> int:
        """:class:`int`: Returns then number of users who are a member of the List.

        .. versionadded:: 1.5.0
        """
        return int(self._payload.get("member_count"))

    @property
    def follower_count(self) -> int:
        """:class:`int`: Returns the number of users who follow the List.

        .. versionadded:: 1.5.0
        """
        return int(self._payload.get("follower_count"))

    @property
    def url(self) -> str:
        """:class:`str`: Returns the list's url.

        .. versionadded:: 1.5.0
        """
        return f"https://twitter.com/i/lists/{self.id}"

    def follow(self) -> RelationFollow:
        """Let the client follows the list.

        Returns
        ---------
        :class:`RelationFollow`
            This method returns a :class:`RelationFollow` object.


        .. versionadded:: 1.5.0
        """
        my_id = self.http_client.access_token.partition("-")[0]
        data = self.http_client.request(
            "POST", "2", f"/users/{my_id}/followed_lists", json={"list_id": str(self.id)}, auth=True
        )
        return RelationFollow(data)

    def unfollow(self) -> RelationFollow:
        """Let the client unfollows the list.

        Returns
        ---------
        :class:`RelationFollow`
            This method returns a :class:`RelationFollow` object.


        .. versionadded:: 1.5.0
        """
        my_id = self.http_client.access_token.partition("-")[0]
        data = self.http_client.request("DELETE", "2", f"/users/{my_id}/followed_lists/{self.id}", auth=True)
        return RelationFollow(data)

    def fetch_followers(self) -> UserPagination:
        """Returns a pagination with a list of :class:`User`s as its content.

        Returns
        ---------
        :class:`TweetPagination`
            This method returns a :class:`TweetPagination` object.


        .. versionadded:: 1.5.0
        """
        params = {
            "expansions": PINNED_TWEET_EXPANSION,
            "user.fields": USER_FIELD,
            "tweet.fields": TWEET_FIELD,
        }

        res = self.http_client.request(
            "GET",
            "2",
            f"/lists/{self.id}/followers",
            params=params,
        )

        if not res:
            return []

        return UserPagination(
            res,
            endpoint_request=f"/lists/{self.id}/followers",
            http_client=self.http_client,
            params=params,
        )

    def fetch_tweets(self) -> TweetPagination:
        """Returns a pagination with a list of :class:`Tweet`s as its content,

        Returns
        ---------
        :class:`TweetPagination`
            This method returns a :class:`TweetPagination` object.


        .. versionadded:: 1.5.0
        """
        params = {
            "expansions": TWEET_EXPANSION,
            "user.fields": USER_FIELD,
            "tweet.fields": TWEET_FIELD,
        }

        res = self.http_client.request(
            "GET",
            "2",
            f"/lists/{self.id}/tweets",
            params=params,
        )

        if not res:
            return []

        return TweetPagination(
            res,
            endpoint_request=f"/lists/{self.id}/tweets",
            http_client=self.http_client,
            params=params,
        )

    def update(
        self,
        *,
        name: Optional[str] = None,
        description: Optional[str] = None,
        private: Optional[bool] = None,
    ) -> Optional[RelationUpdate]:
        """Updates the list.

        Parameters
        -----------
        name: :class:`str`
            The name of the List you wish to update.
        description: :class:`str`
            Description of the List.
        private: :class:`bool`
            Determine whether the List should be private, default to None.


        .. versionadded:: 1.5.0
        """
        return self.http_client.update_list(self.id, name=name, description=description, private=private)

    def delete(self) -> Optional[RelationDelete]:
        """Deletes the list.

        Returns
        ---------
        Optional[:class:`RelationDelete`]
            This method returns a :class:`RelationDelete` object.


        .. versionadded:: 1.5.0
        """
        res = self.http_client.request("DELETE", "2", f"/lists/{self.id}", auth=True)
        return RelationDelete(res)

    def pin(self) -> Optional[RelationPin]:
        """Pins the list.

        Returns
        ---------
        Optional[:class:`RelationPin`]
            This method returns a :class:`RelationPin` object.


        .. versionadded:: 1.5.0
        """
        res = self.http_client.request(
            "POST",
            "2",
            f"/users/{self.owner.id}/pinned_lists",
            auth=True,
            json={"list_id": str(self.id)},
        )

        return RelationPin(res)

    def unpin(self) -> Optional[RelationPin]:
        """Unpins the list.

        Returns
        ---------
        Optional[:class:`RelationPin`]
            This method returns a :class:`RelationPin` object.


        .. versionadded:: 1.5.0
        """
        res = self.http_client.request("DELETE", "2", f"/users/{self.owner.id}/pinned_lists/{self.id}", auth=True)

        return RelationPin(res)

    def add_members(self, *users: User):
        """Adds members to the list.

        Parameters
        ------------
        *users: :class:`User`
            An array of :class:`User` arguments that you wist to add from the list. Example:

            .. code-blocks:: python

                some_list.add_members(user1, user2, user3)


        .. versionadded:: 1.5.0
        """
        executor = self.http_client.thread_manager.create_new_executor(thread_name="add-members-list-method")
        for user in users:
            executor.submit(
                self.http_client.request,
                "POST",
                "2",
                f"/lists/{self.id}/members",
                json={"user_id": str(user.id)},
                auth=True,
            )
        executor.wait_for_futures()

    def remove_members(self, *users: User):
        """Removes members to the list.

        Parameters
        ------------
        *users: :class:`User`
            An array of :class:`User` arguments that you wist to delete from the list. Example:

            .. code-blocks:: python

                some_list.remove_members(user1, user2, user3)


        .. versionadded:: 1.5.0
        """
        executor = self.http_client.thread_manager.create_new_executor(thread_name="remove-members-list-method")
        for user in users:
            executor.submit(
                self.http_client.request,
                "DELETE",
                "2",
                f"/lists/{self.id}/members/{user.id}",
                auth=True,
            )
        executor.wait_for_futures()

    def fetch_members(self) -> Optional[UserPagination]:
        res = self.http_client.request(
            "GET",
            "2",
            f"/lists/{self.id}/members",
            params={
                "expansions": PINNED_TWEET_EXPANSION,
                "user.fields": USER_FIELD,
                "tweet.fields": TWEET_FIELD,
            },
            auth=True,
        )
        if not res:
            return []

        return UserPagination(
            res,
            endpoint_request=f"/lists/{self.id}/member",
            http_client=self.http_client,
            params={
                "expansions": PINNED_TWEET_EXPANSION,
                "user.fields": USER_FIELD,
                "tweet.fields": TWEET_FIELD,
            },
        )


class _CopyList(List):  # An alias for documentation (not interfering with List clas)
    pass



"""
The MIT License (MIT)

Copyright (c) 2021-present UnrealFar & TheGenocides
Copyright (c) 2023-present Sengolda

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import requests
from typing import Optional
from json import decoder


class pytwotException(Exception):
    """This is the base class of all exceptions Raise by pytwot. This inherits :class:`Exception`.

    .. versionadded:: 1.2.0
    """

    def __init__(
        self,
        message: str = None,
    ):
        self.message = message
        super().__init__(self.message)


class APIException(pytwotException):
    """A custom error that will be Raise whenever a request returns an HTTP status code 200. This inherits :class:`pytwotException`.

    .. versionadded:: 1.2.0
    """

    def __init__(
        self,
        response: Optional[requests.models.Response] = None,
        message: str = "No Error Message Provided",
    ):
        self.res = response
        self.message = message
        super().__init__(f"API returned an Exception: {self.message}")


class HTTPException(pytwotException):
    """A custom error that will be Raise whenever a request returns an HTTP status code above 200. This inherits :class:`pytwotException`.

    .. versionadded:: 1.2.0
    """

    def __init__(
        self,
        response: Optional[requests.models.Response] = None,
        message: str = None,
    ):
        self.response = response
        self.message = message
        self.detail = None
        if response is not None:
            try:
                res = self.response.json()
                if res.get("errors"):
                    self.message = res.get("errors")[0].get("message") if not message else message
                    self.detail = res.get("errors")[0].get("detail")

                else:
                    self.message = res.get("error")
                    if not self.message:
                        self.detail = res.get("detail")

            except decoder.JSONDecodeError:
                super().__init__(
                    f"Request returned an Exception (status code: {self.response.status_code}): {self.response.text}",
                )

            else:
                super().__init__(
                    f"Request returned an Exception (status code: {self.response.status_code}): {self.message if self.message else self.detail}",
                )

        else:
            super().__init__(
                f"Exception Raise: {self.message}",
            )

    @property
    def status_code(self) -> Optional[int]:
        if not self.response:
            return None
        return self.response.status_code


class BadRequests(HTTPException):
    """This class inherits :class:`HTTPException`. Raise when a request return status code: 400.

    .. versionadded:: 1.2.0
    """

    pass


class Unauthorized(HTTPException):
    """This class inherits :class:`HTTPException`. Raise when the credentials you passed are invalid and a request returns status code: 401

    .. versionadded:: 1.0.0
    """

    pass


class Forbidden(HTTPException):
    """This class inherits :class:`HTTPException`. Raises when a request returns status code: 403.

    .. versionadded:: 1.2.0
    """

    pass


class FieldsTooLarge(HTTPException):
    """ "This class inherits :class:`HTTPException`. Raises when a request returns status code: 431"""

    pass


class NotFound(HTTPException):
    """This class inherits :class:`HTTPException`. Raise when a request returns status code: 404.

    .. versionadded:: 1.2.0
    """

    pass


class TooManyRequests(HTTPException):
    """This class inherits :class:`HTTPException`. Raise when ratelimit exceeded and a request return status code 429.

    .. versionadded:: 1.1.0
    """

    pass


class ConnectionException(HTTPException):
    """This error class inherits :class:`HTTPException`. This error is Raise when a stream connection throw an error.

    .. versionadded:: 1.3.5
    """

    pass


class Conflict(HTTPException):
    """This error class inherits :class:`HTTPException`. This error is Raise when a request return 409 status code."""

    pass


class UnKnownSpaceState(APIException):
    """This error class inherits :class:`APIException`. This error is Raise when a user specified an invalid space state.

    .. versionadded:: 1.5.0
    """

    def __init__(self, given_state):
        super().__init__(message="Unknown state passed: %s" % given_state)


class NoPageAvailable(APIException):
    """This error class inherits :class:`APIException`. This error is Raise when a user try to lookup a new page in a pagination object that does not exist. These following methods can Raise this error:

    * :meth:`UserPagination.next_page` and :meth:`UserPagination.previous_page`
    * :meth:`TweetPagination.next_page` and :meth:`TweetPagination.previous_page`
    * :meth:`ListPagination.next_page` and :meth:`ListPagination.previous_page`
    * :meth:`MessagePagination.next_page` and :meth:`MessagePagination.previous_page`

    .. versionadded:: 1.5.0
    """

    def __init__(self):
        super().__init__(message="Pagination has no more page available!")


class UnauthorizedForResource(APIException):
    """This error class inherits :class:`APIException`. This error is Raise when the client is unauthorized to view certain resource like for example: viewing a protected user's tweets using :meth:`User.fetch_timelines`.

    .. versionadded:: 1.5.0
    """

    def __init__(self, message):
        super().__init__(message=message)


class ResourceNotFound(APIException):
    """This error class inherits :class:`APIException`. This error is a result of finding a none existent resource .

    .. versionadded:: 1.0.0

    .. versionchanged:: 1.5.0

        Changed name from `NotFoundError` to `ResourceNotFound`
    """

    def __init__(self, message):
        super().__init__(message=message)


class DisallowedResource(APIException):
    """This error class inherits :class:`APIException`. This error is a result of finding a none existent resource .

    .. versionadded:: 1.5.0
    """

    def __init__(self, message):
        super().__init__(message=message)
